# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 6 June 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike
from scipy.special import ellipeinc

from ... import constants as const
from ...mathtime.vector import rot1, rot2, rot3, angle, unit
from ..time.data import IAU80Array
from ..time.frame_conversions import ecef2eci, eci2ecef
from .newton import newtonnu, newtonm, newtone
from . import utils


########################################################################################
# Spherical Elements
########################################################################################


def adbar2rv(
    rmag: float, vmag: float, rtasc: float, decl: float, fpav: float, az: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Conversion from spherical elements to ECI position & velocity vectors.

    References:
        Vallado: 2001, XX
        Chobotov: 2006, p. 70

    Args:
        rmag (float): ECI position vector magnitude in km
        vmag: (float): ECI velocity vector magnitude in km/sec
        rtasc (float): Right ascension of satellite in radians
        decl (float): Declination of satellite in radians
        fpav (float): Satellite flight path angle from vertical in radians
        az (float): Satellite flight path azimuth in radians

    Returns:
        tuple: (r, v)
            r (np.ndarray): ECI position vector
            v (np.ndarray): ECI velocity vector
    """
    # Form position vector
    r = np.array(
        [
            rmag * np.cos(decl) * np.cos(rtasc),
            rmag * np.cos(decl) * np.sin(rtasc),
            rmag * np.sin(decl),
        ]
    )

    # Form velocity vector
    v = np.array(
        [
            vmag
            * (
                np.cos(rtasc)
                * (
                    -np.cos(az) * np.sin(fpav) * np.sin(decl)
                    + np.cos(fpav) * np.cos(decl)
                )
                - np.sin(az) * np.sin(fpav) * np.sin(rtasc)
            ),
            vmag
            * (
                np.sin(rtasc)
                * (
                    -np.cos(az) * np.sin(fpav) * np.sin(decl)
                    + np.cos(fpav) * np.cos(decl)
                )
                + np.sin(az) * np.sin(fpav) * np.cos(rtasc)
            ),
            vmag
            * (np.cos(az) * np.cos(decl) * np.sin(fpav) + np.cos(fpav) * np.sin(decl)),
        ]
    )

    return r, v


def rv2adbar(
    r: ArrayLike, v: ArrayLike
) -> Tuple[float, float, float, float, float, float]:
    """Conversion from position & velocity vectors to spherical elements.

    References:
        Vallado: 2001, xx
        Chobotov: 2006, p. 70

    Args:
        r (array_like): ECI position vector
        v (array_like): ECI velocity vector

    Returns:
        tuple: (rmag, vmag, rtasc, decl, fpav, az)
            rmag (float): ECI position vector magnitude in km
            vmag: (float): ECI velocity vector magnitude in km/sec
            rtasc (float): Right ascension of satellite in radians
            decl (float): Declination of satellite in radians
            fpav (float): Satellite flight path angle from vertical in radians
            az (float): Satellite flight path azimuth in radians
    """
    rmag, vmag = np.linalg.norm(r), np.linalg.norm(v)
    rtemp = np.sqrt(r[0] ** 2 + r[1] ** 2)
    vtemp = np.sqrt(v[0] ** 2 + v[1] ** 2)

    # Right ascension of sateillite
    if rtemp < const.SMALL:
        rtasc = np.arctan2(v[1], v[0]) if vtemp > const.SMALL else 0
    else:
        rtasc = np.arctan2(r[1], r[0])

    # Declination of satellite
    decl = np.arcsin(r[2] / rmag)

    # Flight path angle from vertical
    h = np.cross(r, v)
    fpav = np.arctan2(np.linalg.norm(h), np.dot(r, v))

    # Flight path azimuth
    hcrossr = np.cross(h, r)
    az = np.arctan2(r[0] * hcrossr[1] - r[1] * hcrossr[0], hcrossr[2] * rmag)

    return rmag, vmag, rtasc, decl, fpav, az


########################################################################################
# Classical Elements
########################################################################################


def coe2rv(
    p: float,
    ecc: float,
    incl: float,
    raan: float,
    argp: float,
    nu: float = 0.0,
    arglat: float = 0.0,
    truelon: float = 0.0,
    lonper: float = 0.0,
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert from classical elements to position & velocity vectors.

    References:
        Vallado: 2022, p. 120-121, Algorithm 10
        Chobotov: 2006, p. 70

    Args:
        p (float): Semi-latus rectum of the orbit in km
        ecc (float): Eccentricity of the orbit
        incl (float): Inclination of the orbit in radians
        raan (float): Right ascension of the ascending node (RAAN) in radians
        argp (float): Argument of perigee in radians
        nu (float, optional): True anomaly in radians
        arglat (float, optional): Argument of latitude in radians
        truelon (float, optional): True longitude in radians
        lonper (float, optional): Longitude of periapsis in radians

    Returns:
        tuple: (r, v)
            r (np.ndarray): ECI position vector in km
            v (np.ndarray): ECI velocity vector in km/s
    """
    # Handle special cases for orbit type
    if ecc < const.SMALL:
        if utils.is_equatorial(incl):
            # Circular equatorial
            argp, raan = 0, 0
            nu = truelon
        else:
            # Circular inclined
            argp = 0
            nu = arglat
    else:
        # Elliptical equatorial
        if utils.is_equatorial(incl):
            argp = lonper
            raan = 0

    # Compute position and velocity in the perifocal coordinate system
    cosnu, sinnu = np.cos(nu), np.sin(nu)
    r_pqw = np.array([cosnu, sinnu, 0]) * (p / (1 + ecc * cosnu))
    v_pqw = np.array([-sinnu, ecc + cosnu, 0]) * (np.sqrt(const.MU / p))

    # Transform from PQW to IJK (GEC)
    r = rot3(rot1(rot3(r_pqw, -argp), -incl), -raan)
    v = rot3(rot1(rot3(v_pqw, -argp), -incl), -raan)

    return r, v


def rv2coe(
    r: ArrayLike, v: ArrayLike
) -> Tuple[
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    utils.OrbitType | None,
]:
    """Converts position and velocity vectors into classical orbital elements.

    Args:
        r (array_like): Position vector in km
        v (array_like): Velocity vector in km/s

    References:
        Vallado: 2022, p. 115-116, Algorithm 9

    Returns:
        tuple: (p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper,
                orbit_type)
            p (float): Semilatus rectum in km
            a (float): Semimajor axis in km
            ecc (float): Eccentricity
            incl (float): Inclination in radians (0 to 2pi)
            raan (float): Right ascension of the ascending node in radians
                          (0 to 2pi)
            argp (float): Argument of perigee in radians (0 to 2pi)
            nu (float): True anomaly in radians (0 to 2pi)
            m (float): Mean anomaly in radians (0 to 2pi)
            arglat (float): Argument of latitude in radians (0 to 2pi)
            truelon (float): True longitude in radians (0 to 2pi)
            lonper (float): Longitude of periapsis in radians (0 to 2pi)
            orbit_type (enum): Type of orbit as defined in the OrbitType enum
    """

    def adjust_angle(ang):
        """Adjust angle by subtracting it from 2pi"""
        return const.TWOPI - ang

    # Initialize variables
    p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper = (np.nan,) * 11
    orbit_type = None

    # Make sure position and velocity vectors are numpy arrays
    r, v = np.array(r), np.array(v)

    # Get magnitude of position and velocity vectors
    r_mag, v_mag = np.linalg.norm(r), np.linalg.norm(v)

    # Get angular momentum
    h = np.cross(r, v)
    h_mag = np.linalg.norm(h)

    # Elements are undefined for negative angular momentum
    if h_mag < 0:
        return p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper, orbit_type

    # Define line of nodes vector
    n_vec = np.array([-h[1], h[0], 0])
    n_mag = np.linalg.norm(n_vec)

    # Get eccentricity vector
    e_vec = ((v_mag**2 - const.MU / r_mag) * r - np.dot(r, v) * v) / const.MU
    ecc = np.linalg.norm(e_vec)

    # find a, e, and p (semi-latus rectum)
    sme = (v_mag**2 / 2) - (const.MU / r_mag)
    a = -const.MU / (2 * sme) if abs(sme) > const.SMALL else np.inf

    # Semi-latus rectum
    p = h_mag**2 / const.MU

    # Find inclination
    incl = np.arccos(h[2] / h_mag)

    # Determine orbit type
    orbit_type = utils.determine_orbit_type(ecc, incl, tol=const.SMALL)

    # Find right ascension of ascending node
    if n_mag > const.SMALL:
        raan = np.arccos(np.clip(n_vec[0] / n_mag, -1, 1))
        if n_vec[1] < 0:
            raan = adjust_angle(raan)

    # Find argument of periapsis
    if orbit_type is utils.OrbitType.EPH_INCLINED:
        argp = angle(n_vec, e_vec)
        if e_vec[2] < 0:
            argp = adjust_angle(argp)

    # Find true anomaly at epoch
    if orbit_type in [utils.OrbitType.EPH_INCLINED, utils.OrbitType.EPH_EQUATORIAL]:
        nu = angle(e_vec, r)
        if np.dot(r, v) < 0:
            nu = adjust_angle(nu)

    # Find argument of latitude (inclined cases)
    if orbit_type in [utils.OrbitType.CIR_INCLINED, utils.OrbitType.EPH_INCLINED]:
        arglat = angle(n_vec, r)
        if r[2] < 0:
            arglat = adjust_angle(arglat)

    # Find longitude of periapsis
    if ecc > const.SMALL and orbit_type is utils.OrbitType.EPH_EQUATORIAL:
        lonper = np.arccos(np.clip(e_vec[0] / ecc, -1, 1))
        if e_vec[1] < 0:
            lonper = adjust_angle(lonper)
        if incl > np.pi / 2:
            lonper = adjust_angle(lonper)

    # Find true longitude
    if r_mag > const.SMALL and orbit_type is utils.OrbitType.CIR_EQUATORIAL:
        truelon = np.arccos(np.clip(r[0] / r_mag, -1, 1))
        if r[1] < 0:
            truelon = adjust_angle(truelon)
        if incl > np.pi / 2:
            truelon = adjust_angle(truelon)

    # Find mean anomaly for eccentric orbits
    if np.isreal(ecc):
        _, m = newtonnu(ecc, nu)

    return p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper, orbit_type


########################################################################################
# Equinoctial Elements
########################################################################################


def eq2rv(
    a: float, af: float, ag: float, chi: float, psi: float, meanlon: float, fr: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert from equinoctial elements to position & velocity vectors.

    This function finds the position and velocity vectors in geocentric equatorial (ijk)
    system given the equinoctial orbit elements.

    References:
        Vallado: 2022, p. 110-111

    Args:
        a (float): Semi-major axis in km
        af (float): Component of eccentricity vector (also called k)
        ag (float): Component of eccentricity vector (also called h)
        chi (float): Component of node vector in eqw (also called p)
        psi (float): Component of node vector in eqw (also called q)
        meanlon (float): Mean longitude in radians
        fr (int): Retrograde factor (+1 for prograde, -1 for retrograde)

    Returns:
        tuple: (r, v)
            r (np.ndarray): ECI position vector in km
            v (np.ndarray): ECI velocity vector in km/s

    TODO:
        - Add vector option for conversion
    """
    # Initialize variables
    arglat, truelon, lonper = (0,) * 3

    # Compute eccentricity
    ecc = np.sqrt(af**2 + ag**2)
    p = a * (1 - ecc**2)
    incl = np.pi * ((1 - fr) * 0.5) + 2 * fr * np.arctan(np.sqrt(chi**2 + psi**2))
    raan = np.arctan2(chi, psi)
    argp = np.arctan2(ag, af) - fr * raan

    if ecc < const.SMALL:
        # Circular orbits
        if utils.is_equatorial(incl):
            # Circular equatorial
            truelon = raan
            raan = 0
        else:
            # Circular inclined
            arglat = argp
    else:
        # Elliptical equatorial
        if utils.is_equatorial(incl):
            lonper = argp
            raan = 0

    # Mean anomaly
    m = meanlon - fr * raan - argp
    m = np.mod(m, const.TWOPI)

    # Solve for eccentric anomaly and true anomaly
    e0, nu = newtonm(ecc, m)

    if ecc < const.SMALL:
        # Circular orbits
        if utils.is_equatorial(incl):
            # Circular equatorial
            truelon = nu
        else:
            # Circular inclined
            arglat = nu - fr * raan

    # Convert back to position and velocity vectors
    return coe2rv(p, ecc, incl, raan, argp, nu, arglat, truelon, lonper)


def rv2eq(
    r: ArrayLike, v: ArrayLike
) -> Tuple[float, float, float, float, float, float, float, float, int]:
    """Convert from position & velocity vectors to equinoctial elements.

    References:
        Vallado: 2022, p. 110-111
        Chobotov: 2006, p. 30

    Args:
        r (array_like): ECI position vector in km
        v (array_like): ECI velocity vector in km/s

    Returns:
        tuple: (a, n, af, ag, chi, psi, meanlon, truelon, fr)
            a (float): Semi-major axis in km
            n (float): Mean motion in rad/s
            af (float): Component of eccentricity vector
            ag (float): Component of eccentricity vector
            chi (float): Component of node vector in eqw
            psi (float): Component of node vector in eqw
            meanlon (float): Mean longitude in radians
            meanlonNu (float): True longitude in radians
            fr (int): Retrograde factor (+1 for prograde, -1 for retrograde)
    """
    # Convert to classical orbital elements
    p, a, ecc, incl, omega, argp, nu, m, arglat, truelon, lonper, _ = rv2coe(r, v)

    # Determine retrograde factor
    fr = -1 if abs(incl - np.pi) < const.SMALL else 1

    if ecc < const.SMALL:
        # Circular orbits
        if utils.is_equatorial(incl):
            # Circular Equatorial
            argp, omega = 0, 0
            nu, m = truelon, truelon
        else:
            # Circular inclined
            argp = 0
            nu, m = arglat
    else:
        # Elliptical equatorial
        if utils.is_equatorial(incl):
            argp = lonper
            omega = 0

    # Calculate mean motion
    # TODO: put in separate utility function
    n = np.sqrt(const.MU / (a**3))

    # Get eccentricity vector components
    af = ecc * np.cos(fr * omega + argp)
    ag = ecc * np.sin(fr * omega + argp)

    # Get EQW node vector components
    if fr > 0:
        chi = np.tan(incl * 0.5) * np.sin(omega)
        psi = np.tan(incl * 0.5) * np.cos(omega)
    else:
        chi = 1 / np.tan(incl * 0.5) * np.sin(omega)
        psi = 1 / np.tan(incl * 0.5) * np.cos(omega)

    # Determine mean longitude
    meanlon = fr * omega + argp + m
    meanlon = np.mod(meanlon + const.TWOPI, const.TWOPI)

    # Determine true longitude
    truelon = fr * omega + argp + nu
    truelon = np.mod(truelon + const.TWOPI, const.TWOPI)

    return a, n, af, ag, chi, psi, meanlon, truelon, fr


########################################################################################
# Topocentric Elements
########################################################################################


def tradec2rv(
    trr: float,
    trtasc: float,
    tdecl: float,
    dtrr: float,
    tdrtasc: float,
    tddecl: float,
    rseci: ArrayLike,
    vseci: ArrayLike,
) -> Tuple[np.ndarray, np.ndarray]:
    """Converts topocentric coordinates (range, right ascension, declination, and their
    rates) into geocentric equatorial (ECI) position and velocity vectors.

    References:
        Vallado: 2022, p. 257, Algorithm 26

    Args:
        trr (float): Satellite range from site in km
        trtasc (float): Topocentric right ascension in radians
        tdecl (float): Topocentric declination in radians
        dtrr (float): Range rate in km/s
        tdrtasc (float): Topocentric right ascension rate in rad/s
        tddecl (float): Topocentric declination rate in rad/s
        rseci (array_like): ECI site position vector in km
        vseci (array_like): ECI site velocity vector in km/s

    Returns:
        tuple: (reci, veci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
    """
    # Calculate topocentric slant range vectors
    rhov = np.array(
        [
            trr * np.cos(tdecl) * np.cos(trtasc),
            trr * np.cos(tdecl) * np.sin(trtasc),
            trr * np.sin(tdecl),
        ]
    )

    # Slant range rate vectors
    drhov = np.array(
        [
            dtrr * np.cos(tdecl) * np.cos(trtasc)
            - trr * np.sin(tdecl) * np.cos(trtasc) * tddecl
            - trr * np.cos(tdecl) * np.sin(trtasc) * tdrtasc,
            dtrr * np.cos(tdecl) * np.sin(trtasc)
            - trr * np.sin(tdecl) * np.sin(trtasc) * tddecl
            + trr * np.cos(tdecl) * np.cos(trtasc) * tdrtasc,
            dtrr * np.sin(tdecl) + trr * np.cos(tdecl) * tddecl,
        ]
    )

    # ECI position and velocity vectors
    reci = rhov + rseci
    veci = drhov + vseci

    return reci, veci


def rv2tradec(
    reci: ArrayLike, veci: ArrayLike, rseci: ArrayLike, vseci: ArrayLike
) -> Tuple[float, float, float, float, float, float]:
    """Converts geocentric equatorial (ECI) position and velocity vectors into range,
    topocentric right ascension, declination, and rates.

    References:
        Vallado: 2022, p. 257, Algorithm 26

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like)): ECI velocity vector in km/s
        rseci (array_like)): ECI site position vector in km
        vseci (array_like)): ECI site velocity vector in km/s

    Returns:
        tuple: (rho, trtasc, tdecl, drho, dtrtasc, dtdecl)
            rho (float): Satellite range from site in km
            trtasc (float): Topocentric right ascension in radians (0 to 2pi)
            tdecl (float): Topocentric declination in radians (-pi/2 to pi/2)
            drho (float): Range rate in km/s
            dtrtasc (float): Topocentric right ascension rate in rad/s
            dtdecl (float): Topocentric declination rate in rad/s
    """
    # Find ECI slant range vector from site to satellite
    rhoveci = np.array(reci) - np.array(rseci)
    drhoveci = np.array(veci) - np.array(vseci)
    rho = np.linalg.norm(rhoveci)

    # Calculate topocentric right ascension and declination
    temp = np.sqrt(rhoveci[0] ** 2 + rhoveci[1] ** 2)
    if temp < const.SMALL:
        trtasc = np.arctan2(drhoveci[1], drhoveci[0])
    else:
        trtasc = np.arctan2(rhoveci[1], rhoveci[0])

    # Directly over the North Pole
    if temp < const.SMALL:
        tdecl = np.sign(rhoveci[2]) * np.pi / 2  # +- 90 deg
    else:
        tdecl = np.arcsin(rhoveci[2] / rho)

    if trtasc < 0:
        trtasc += const.TWOPI

    # Calculate topocentric right ascension and declination rates
    temp1 = -rhoveci[1] ** 2 - rhoveci[0] ** 2
    drho = np.dot(rhoveci, drhoveci) / rho
    dtrtasc, dtdecl = 0, 0
    if abs(temp1) > const.SMALL:
        dtrtasc = (drhoveci[0] * rhoveci[1] - drhoveci[1] * rhoveci[0]) / temp1
    if abs(temp) > const.SMALL:
        dtdecl = (drhoveci[2] - drho * np.sin(tdecl)) / temp

    return rho, trtasc, tdecl, drho, dtrtasc, dtdecl


########################################################################################
# Flight Elements
########################################################################################


def flt2rv(
    rmag: float,
    vmag: float,
    latgc: float,
    lon: float,
    fpa: float,
    az: float,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """Converts flight elements into ECI position and velocity vectors.

    References:
        Vallado: 2022, p. 111-112
        Escobal: 1985, p. 397
        Chobotov: 2006, p. 67

    Args:
        rmag (float): Position vector magnitude in km
        vmag (float): Velocity vector magnitude in km/s
        latgc (float): Geocentric latitude in radians
        lon (float): Longitude in radians
        fpa (float): Flight path angle in radians
        az (float): Flight path azimuth in radians
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (reci, veci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
    """
    # Form position vector
    recef = np.array(
        [
            rmag * np.cos(latgc) * np.cos(lon),
            rmag * np.cos(latgc) * np.sin(lon),
            rmag * np.sin(latgc),
        ]
    )

    # Convert r to ECI
    vecef = np.zeros(3)  # this is a dummy for now
    aecef = np.zeros(3)
    reci, veci, _ = ecef2eci(
        recef, vecef, aecef, ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms
    )

    # Calculate right ascension and declination
    if np.sqrt(reci[0] ** 2 + reci[1] ** 2) < const.SMALL:
        rtasc = np.arctan2(veci[1], veci[0])
    else:
        rtasc = np.arctan2(reci[1], reci[0])
    decl = np.arcsin(reci[2] / rmag)

    # Form velocity vector
    fpav = const.HALFPI - fpa
    veci = vmag * np.array(
        [
            # First element
            (
                -np.cos(rtasc)
                * np.sin(decl)
                * (
                    np.cos(az) * np.cos(fpav)
                    - np.sin(rtasc) * np.sin(az) * np.cos(fpav)
                )
                + np.cos(rtasc) * np.sin(decl) * np.sin(fpav)
            ),
            # Second element
            (
                -np.sin(rtasc)
                * np.sin(decl)
                * (
                    np.cos(az) * np.cos(fpav)
                    + np.cos(rtasc) * np.sin(az) * np.cos(fpav)
                )
                + np.sin(rtasc) * np.cos(decl) * np.sin(fpav)
            ),
            # Third element
            (np.sin(decl) * np.sin(fpav) + np.cos(decl) * np.cos(az) * np.cos(fpav)),
        ]
    )

    return reci, veci


def rv2flt(
    reci: ArrayLike,
    veci: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[float, float, float, float, float, float, float, float]:
    """Transforms a position and velocity vector to flight elements.

    References:
        Vallado: 2022, p. 111-112

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (lon, latgc, rtasc, decl, fpa, az, rmag, vmag)
            lon (float): Longitude in radians
            latgc (float): Geocentric latitude in radians
            rtasc (float): Right ascension angle in radians
            decl (float): Declination angle in radians
            fpa (float): Flight path angle in radians
            az (float): Flight path azimuth in radians
            rmag (float): Position vector magnitude in km
            vmag (float): Velocity vector magnitude in km/s
    """
    # Get magnitude of position and velocity vectors
    rmag, vmag = np.linalg.norm(reci), np.linalg.norm(veci)

    # Convert r to ECEF for lat/lon calculations
    aecef = np.zeros(3)
    recef, vecef, aecef = eci2ecef(
        reci, veci, aecef, ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms
    )

    # Calculate longitude
    if np.sqrt(recef[0] ** 2 + recef[1] ** 2) < const.SMALL:
        lon = np.arctan2(vecef[1], vecef[0])
    else:
        lon = np.arctan2(recef[1], recef[0])

    latgc = np.arcsin(recef[2] / rmag)

    # Calculate right ascension and declination
    if np.sqrt(reci[0] ** 2 + reci[1] ** 2) < const.SMALL:
        rtasc = np.arctan2(veci[1], veci[0])
    else:
        rtasc = np.arctan2(reci[1], reci[0])

    decl = np.arcsin(reci[2] / rmag)

    # Calculate flight path angle
    h = np.cross(reci, veci)
    hmag = np.linalg.norm(h)
    rdotv = np.dot(reci, veci)
    fpav = np.arctan2(hmag, rdotv)
    fpa = const.HALFPI - fpav

    # Calculte azimuth
    hcrossr = np.cross(h, reci)
    az = np.arctan2(reci[0] * hcrossr[1] - reci[1] * hcrossr[0], hcrossr[2] * rmag)

    return lon, latgc, rtasc, decl, fpa, az, rmag, vmag


########################################################################################
# Ecliptic Elements
########################################################################################


def ell2rv(
    rr: float, ecllon: float, ecllat: float, drr: float, decllon: float, decllat: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms ecliptic latitude and longitude to position and velocity vectors.

    References:
        Vallado: 2022, p. 265-267

    Args:
        rr (float): Radius of the satellite in km
        ecllon (float): Ecliptic longitude in radians
        ecllat (float): Ecliptic latitude in radians
        drr (float): Radius of the satellite rate in km/s
        decllon (float): Ecliptic longitude rate of change in rad/s
        decllat (float): Ecliptic latitude rate of change in rad/s

    Returns:
        tuple: (reci, veci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
    """
    # Calculate position vector in ecliptic coordinates
    r = np.array(
        [
            rr * np.cos(ecllat) * np.cos(ecllon),
            rr * np.cos(ecllat) * np.sin(ecllon),
            rr * np.sin(ecllat),
        ]
    )

    # Calculate velocity vector in ecliptic coordinates
    v = np.array(
        [
            # X component
            drr * np.cos(ecllat) * np.cos(ecllon)
            - rr * np.sin(ecllat) * np.cos(ecllon) * decllat
            - rr * np.cos(ecllat) * np.sin(ecllon) * decllon,
            # Y component
            drr * np.cos(ecllat) * np.sin(ecllon)
            - rr * np.sin(ecllat) * np.sin(ecllon) * decllat
            + rr * np.cos(ecllat) * np.cos(ecllon) * decllon,
            # Z component
            drr * np.sin(ecllat) + rr * np.cos(ecllat) * decllat,
        ]
    )

    # Rotate position and velocity vectors to the ECI frame
    reci = rot1(r, -const.OBLIQUITYEARTH)
    veci = rot1(v, -const.OBLIQUITYEARTH)

    return reci, veci


def rv2ell(
    reci: ArrayLike, veci: ArrayLike
) -> Tuple[float, float, float, float, float, float]:
    """Transforms position and velocity vectors to ecliptic latitude and longitude.

    References:
        Vallado: 2022, p. 265-267

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s

    Returns:
        tuple: (rr, ecllon, ecllat, drr, decllon, decllat)
            rr (float): Radius of the satellite in km
            ecllon (float): Ecliptic longitude in radians
            ecllat (float): Ecliptic latitude in radians
            drr (float): Radius of the satellite rate in km/s
            decllon (float): Ecliptic longitude rate of change in rad/s
            decllat (float): Ecliptic latitude rate of change in rad/s
    """
    # Perform rotation about the x-axis by the obliquity
    r = rot1(reci, const.OBLIQUITYEARTH)
    v = rot1(veci, const.OBLIQUITYEARTH)

    # Calculate magnitudes
    rr = np.linalg.norm(r)
    temp = np.sqrt(r[0] ** 2 + r[1] ** 2)

    # Calculate ecliptic longitude
    if temp < const.SMALL:
        temp1 = np.sqrt(v[0] ** 2 + v[1] ** 2)
        ecllon = np.arctan2(v[1], v[0]) if abs(temp1) > const.SMALL else 0
    else:
        ecllon = np.arctan2(r[1], r[0])

    # Calculate ecliptic latitude
    ecllat = np.arcsin(r[2] / rr)

    # Calculate rates
    temp1 = -r[1] ** 2 - r[0] ** 2  # Different now
    drr = np.dot(r, v) / rr
    decllon = (v[0] * r[1] - v[1] * r[0]) / temp1 if abs(temp1) > const.SMALL else 0
    decllat = (v[2] - drr * np.sin(ecllat)) / temp if abs(temp) > const.SMALL else 0

    return rr, ecllon, ecllat, drr, decllon, decllat


########################################################################################
# Celestial Elements
########################################################################################


def radec2rv(
    rr: float, rtasc: float, decl: float, drr: float, drtasc: float, ddecl: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms celestial (right ascension and declination) elements to position and
    velocity vectors.

    References:
        Vallado: 2022, p. 254-256, Algorithm 25

    Args:
        rr (float): Radius of the satellite in km
        rtasc (float): Right ascension in radians
        decl (float): Declination in radians
        drr (float): Radius of the satellite rate in km/s
        drtasc (float): Right ascension rate in rad/s
        ddecl (float): Declination rate in rad/s

    Returns:
        tuple: (r, v)
            r (np.ndarray): ECI position vector in km
            v (np.ndarray): ECI velocity vector in km/s
    """
    # Position vector
    r = np.array(
        [
            rr * np.cos(decl) * np.cos(rtasc),
            rr * np.cos(decl) * np.sin(rtasc),
            rr * np.sin(decl),
        ]
    )

    # Velocity vector
    v = np.array(
        [
            # X component
            drr * np.cos(decl) * np.cos(rtasc)
            - rr * np.sin(decl) * np.cos(rtasc) * ddecl
            - rr * np.cos(decl) * np.sin(rtasc) * drtasc,
            # Y component
            drr * np.cos(decl) * np.sin(rtasc)
            - rr * np.sin(decl) * np.sin(rtasc) * ddecl
            + rr * np.cos(decl) * np.cos(rtasc) * drtasc,
            # Z component
            drr * np.sin(decl) + rr * np.cos(decl) * ddecl,
        ]
    )

    return r, v


def rv2radec(
    r: ArrayLike, v: ArrayLike
) -> Tuple[float, float, float, float, float, float]:
    """Transforms position and velocity vectors to celestial (right ascension and
    declination) elements.

    References:
        Vallado: 2022, p. 254-256, Algorithm 25

    Args:
        r (array_like): ECI position vector in km
        v (array_like): ECI velocity vector in km/s

    Returns:
        tuple: (rr, rtasc, decl, drr, drtasc, ddecl)
            rr (float): Radius of the satellite in km
            rtasc (float): Right ascension in radians
            decl (float): Declination in radians
            drr (float): Radius of the satellite rate in km/s
            drtasc (float): Right ascension rate in rad/s
            ddecl (float): Declination rate in rad/s
    """
    # Calculate the magnitude of the position vector
    rr = np.linalg.norm(r)
    temp = np.sqrt(r[0] ** 2 + r[1] ** 2)

    # Calculate right ascension
    rtasc = np.arctan2(v[1], v[0]) if temp < const.SMALL else np.arctan2(r[1], r[0])
    rtasc += const.TWOPI if rtasc < 0 else rtasc

    # Calculate declination
    decl = np.arcsin(r[2] / rr)

    # Calculate radius rate
    drr = np.dot(r, v) / rr
    temp1 = -r[1] * r[1] - r[0] * r[0]

    # Calculate right ascension rate
    drtasc = (v[0] * r[1] - v[1] * r[0]) / temp1 if abs(temp1) > const.SMALL else 0

    # Calculate declination rate
    ddecl = (v[2] - drr * np.sin(decl)) / temp if abs(temp) > const.SMALL else 0

    return rr, rtasc, decl, drr, drtasc, ddecl


########################################################################################
# Azimuth-Elevation Elements
########################################################################################


def razel2rvsez(
    rho: float, az: float, el: float, drho: float, daz: float, del_el: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Converts range, azimuth, and elevation values with slant range and velocity
    vectors for a satellite from a radar site in the topocentric horizon (SEZ) system.

    References:
        Vallado: 2022, p. 258-259, Eqs. 4-4 and 4-5

    Args:
        rho (float): Satellite range from site in km
        az (float): Azimuth in radians (0 to 2pi)
        el (float): Elevation in radians (-pi/2 to pi/2)
        drho (float): Range rate in km/s
        daz (float): Azimuth rate in rad/s
        del_el (float): Elevation rate in rad/s

    Returns:
        tuple: (rhosez, drhosez)
            rhosez (np.ndarray): SEZ range vector in km
            drhosez (np.ndarray): SEZ velocity vector in km/s
    """
    # Initialize values
    sinel, cosel = np.sin(el), np.cos(el)
    sinaz, cosaz = np.sin(az), np.cos(az)

    # Form SEZ range vector
    rhosez = np.array([-rho * cosel * cosaz, rho * cosel * sinaz, rho * sinel])

    # Form SEZ velocity vector
    drhosez = np.array(
        [
            -drho * cosel * cosaz + rhosez[2] * del_el * cosaz + rhosez[1] * daz,
            drho * cosel * sinaz - rhosez[2] * del_el * sinaz - rhosez[0] * daz,
            drho * sinel + rho * del_el * cosel,
        ]
    )

    return rhosez, drhosez


def rvsez2razel(
    rhosez: ArrayLike, drhosez: ArrayLike
) -> Tuple[float, float, float, float, float, float]:
    """Transforms SEZ range and velocity vectors to range, azimuth, and elevation values
    and their rates.

    References:
        Vallado: 2022, p. 259-263, Algorithm 27

    Args:
        rhosez (array_like): SEZ range vector in km
        drhosez (array_like): SEZ velocity vector in km/s

    Returns:
        tuple: (rho, az, el, drho, daz, del_el)
            rho (float): Satellite range from site in km
            az (float): Azimuth in radians (0 to 2pi)
            el (float): Elevation in radians (-pi/2 to pi/2)
            drho (float): Range rate in km/s
            daz (float): Azimuth rate in rad/s
            del_el (float): Elevation rate in rad/s
    """
    # Range magnitude
    rho = np.linalg.norm(rhosez)

    # Calculate azimuth
    temp = np.sqrt(rhosez[0] ** 2 + rhosez[1] ** 2)
    if abs(rhosez[1]) < const.SMALL:
        if temp < const.SMALL:
            az = np.arctan2(drhosez[1], -drhosez[0])
        else:
            az = np.pi if rhosez[0] > 0 else 0
    else:
        az = np.arctan2(rhosez[1], -rhosez[0])

    # Calculate elevation
    el = (
        np.sign(rhosez[2]) * const.HALFPI
        if temp < const.SMALL
        else np.arcsin(rhosez[2] / rho)
    )

    # Range rate
    drho = np.dot(rhosez, drhosez) / rho

    # Azimuth rate
    daz = (
        (drhosez[0] * rhosez[1] - drhosez[1] * rhosez[0]) / (temp**2)
        if abs(temp**2) > const.SMALL
        else 0
    )

    # Elevation rate
    del_el = (drhosez[2] - drho * np.sin(el)) / temp if abs(temp) > const.SMALL else 0

    return rho, az, el, drho, daz, del_el


def razel2rv(
    rho: float,
    az: float,
    el: float,
    drho: float,
    daz: float,
    del_el: float,
    latgd: float,
    lon: float,
    alt: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms range, azimuth, elevation, and their rates to the geocentric
    equatorial (ECI) position and velocity vectors.

    References:
        Vallado: 2022, p. 259-263, Algorithm 27

    Args:
        rho (float): Satellite range from site in km
        az (float): Azimuth in radians
        el (float): Elevation in radians
        drho (float): Range rate in km/s
        daz (float): Azimuth rate in rad/s
        del_el (float): Elevation rate in rad/s
        latgd (float): Geodetic latitude of site in radians
        lon (float): Longitude of site in radians
        alt (float): Altitude of site in km

    Returns:
        tuple: (recef, vecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
    """
    # Find SEZ range and velocity vectors
    rhosez, drhosez = razel2rvsez(rho, az, el, drho, daz, del_el)

    # Perform SEZ to ECEF transformation
    rhoecef = rot3(rot2(rhosez, latgd - const.HALFPI), -lon).T
    drhoecef = rot3(rot2(drhosez, latgd - const.HALFPI), -lon).T

    # Find ECEF range and velocity vectors
    rs, vs = utils.site(latgd, lon, alt)
    recef = rhoecef + rs
    vecef = drhoecef

    return recef, vecef


def rv2razel(
    recef: ArrayLike, vecef: ArrayLike, latgd: float, lon: float, alt: float
) -> Tuple[float, float, float, float, float, float]:
    """Transforms ECEF position and velocity vectors to range, azimuth, elevation, and
    their rates.

    The value of `SMALL` can affect the rate term calculations. the solution uses the
    velocity vector to find the singular cases. also, the elevation and azimuth rate
    terms are not observable unless the acceleration vector is available.

    References:
        Vallado: 2022, p. 259-263, Algorithm 27

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        latgd (float): Geodetic latitude of site in radians
        lon (float): Longitude of site in radians
        alt (float): Altitude of site in km

    Returns:
        tuple: (rho, az, el, drho, daz, del_el)
            rho (float): Satellite range from site in km
            az (float): Azimuth in radians (0 to 2pi)
            el (float): Elevation in radians (-pi/2 to pi/2)
            drho (float): Range rate in km/s
            daz (float): Azimuth rate in rad/s
            del_el (float): Elevation rate in rad/s
    """
    # Get site vector in ECEF
    rsecef, _ = utils.site(latgd, lon, alt)

    # Find ECEF range vector from site to satellite
    rhoecef = recef - rsecef
    drhoecef = vecef
    rho = np.linalg.norm(rhoecef)

    # Convert to SEZ for calculations
    tempvec = rot3(rhoecef, lon)
    rhosez = rot2(tempvec, const.HALFPI - latgd)

    tempvec = rot3(drhoecef, lon)
    drhosez = rot2(tempvec, const.HALFPI - latgd)

    # Calculate azimuth and elevation
    temp = np.sqrt(rhosez[0] ** 2 + rhosez[1] ** 2)
    if temp < const.SMALL:
        el = np.sign(rhosez[2]) * const.HALFPI
        az = np.arctan2(drhosez[1], -drhosez[0])
    else:
        magrhosez = np.linalg.norm(rhosez)
        el = np.arcsin(rhosez[2] / magrhosez)
        az = np.arctan2(rhosez[1] / temp, -rhosez[0] / temp)

    # Calculate range, azimuth, and elevation rates
    drho = np.dot(rhosez, drhosez) / rho
    daz = (
        (drhosez[0] * rhosez[1] - drhosez[1] * rhosez[0]) / (temp * temp)
        if temp > const.SMALL
        else 0
    )
    del_el = (drhosez[2] - drho * np.sin(el)) / temp if abs(temp) > const.SMALL else 0

    return rho, az, el, drho, daz, del_el


########################################################################################
# Site Topocentric (SEZ) Elements
########################################################################################


def _sez_transformation_matrix(lat: float, lon: float) -> np.ndarray:
    """Computes the transformation matrix from ECI to SEZ."""
    # Zenith component
    zvec = unit(
        np.array([np.cos(lat) * np.cos(lon), np.cos(lat) * np.sin(lon), np.sin(lat)])
    )

    # East component
    kvec = np.array([0, 0, 1])
    evec = unit(np.cross(kvec, zvec))

    # South component
    svec = unit(np.cross(evec, zvec))

    # Transformation matrix
    return np.vstack([svec, evec, zvec])


def rv2sez(
    reci: ArrayLike, veci: ArrayLike, lat: float, lon: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Converts position and velocity vectors into site topocentric (SEZ) coordinates.

    References:
        Vallado: 2022, p. 165-166

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s
        lat (float): Relative latitude of the second satellite w.r.t. the first
                     in radians
        lon (float): Relative longitude of the second satellite w.r.t the first
                     in radians

    Returns:
        tuple: (rsez, vsez, transmat)
            rsez (np.ndarray): SEZ position vector in km
            vsez (np.ndarray): SEZ velocity vector in km/s
            transmat (np.ndarray): Transformation matrix from ECI to SEZ
    """
    # Transformation matrix from ECI to SEZ
    transmat = _sez_transformation_matrix(lat, lon)

    # Transform position and velocity vectors
    rsez = transmat @ reci
    vsez = transmat @ veci

    return rsez, vsez, transmat


def sez2rv(
    rsez: ArrayLike, vsez: ArrayLike, lat: float, lon: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Converts site topocentric (SEZ) coordinates into ECI position and velocity
    vectors.

    References:
        Vallado: 2022, p. 165-166

    Args:
        rsez (array_like): SEZ position vector in km
        vsez (array_like): SEZ velocity vector in km/s
        lat (float): Relative latitude of the second satellite w.r.t. the first
                     in radians
        lon (float): Relative longitude of the second satellite w.r.t the first
                     in radians

    Returns:
        tuple: (reci, veci, transmat)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            transmat (np.ndarray): Transformation matrix from SEZ to ECI
    """
    # Transformation matrix from SEZ to ECI
    transmat = _sez_transformation_matrix(lat, lon).T

    # Transform back to ECI
    reci = transmat @ rsez
    veci = transmat @ vsez

    return reci, veci, transmat


########################################################################################
# Satellite Coordinate Systems
########################################################################################


def rv2rsw(
    reci: ArrayLike, veci: ArrayLike
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms position and velocity vectors into radial, tangential (in-track), and
    normal (cross-track) coordinates, i.e. RSW frame.

    Note: There are numerous nomenclatures for these systems. This is the RSW system of
    Vallado. The reverse values are found using the transmat transpose.

    References:
        Vallado: 2022, p. 166

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s

    Returns:
        tuple: (rrsw, vrsw, transmat)
            rrsw (np.ndarray): RSW position vector in km
            vrsw (np.ndarray): RSW velocity vector in km/s
            transmat (np.ndarray): Transformation matrix from ECI to RSW
    """
    # Radial component
    rvec = unit(reci)

    # Cross-track component
    wvec = unit(np.cross(reci, veci))

    # Along-track component
    svec = unit(np.cross(wvec, rvec))

    # Assemble transformation matrix from ECI to RSW frame
    transmat = np.vstack([rvec, svec, wvec])

    rrsw = np.dot(transmat, reci)
    vrsw = np.dot(transmat, veci)

    return rrsw, vrsw, transmat


def rv2ntw(
    reci: ArrayLike, veci: ArrayLike
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms position and velocity vectors into normal (in-radial), tangential
    (velocity), and normal (cross-track) coordinates.

    Note: Sometimes the first vector is called "along-radial". the tangential direction
    is always aligned with the velocity vector. This is the NTW system of Vallado.

    References:
        Vallado: 2022, p. 166

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s

    Returns:
        tuple: (rntw, vntw, transmat)
            rntw (np.ndarray): NTW position vector in km
            vntw (np.ndarray): NTW velocity vector in km/s
            transmat (np.ndarray): Transformation matrix from ECI to NTW
    """
    # In-velocity component
    tvec = unit(veci)

    # Cross-track component
    wvec = unit(np.cross(reci, veci))

    # Along-radial component
    nvec = unit(np.cross(tvec, wvec))

    # Assemble transformation matrix from ECI to NTW frame
    transmat = np.array([nvec, tvec, wvec])

    rntw = np.dot(transmat, reci)
    vntw = np.dot(transmat, veci)

    return rntw, vntw, transmat


def rv2pqw(reci: ArrayLike, veci: ArrayLike) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms position and velocity vectors into perifocal (PQW) coordinates.

    References:
        Vallado: 2022, p. 166

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s

    Returns:
        tuple: (rpqw, vpqw)
            rpqw (np.ndarray): PQW position vector in km
            vpqw (np.ndarray): PQW velocity vector in km/s
    """
    # Covert to classical elements
    p, _, ecc, _, _, _, nu, *_ = rv2coe(reci, veci)

    # Return if true anomaly is undefined
    if np.isnan(nu):
        return np.zeros(3), np.zeros(3)

    # Form PQW position vector
    sinnu, cosnu = np.sin(nu), np.cos(nu)
    temp = p / (1 + ecc * cosnu)
    rpqw = temp * np.array([cosnu, sinnu, 0])

    # Form PQW velocity vector
    p = const.SMALL if np.abs(p) < const.SMALL else p
    vpqw = np.array(
        [-np.sqrt(const.MU / p) * sinnu, np.sqrt(const.MU / p) * (ecc + cosnu), 0]
    )

    return rpqw, vpqw


########################################################################################
# Geodetic Elements
########################################################################################


def ecef2ll(r: ArrayLike) -> Tuple[float, float, float, float]:
    """Converts an ECEF position vector into geodetic and geocentric latitude,
    longitude, and height above the ellipsoid using the Astronomical Almanac method.

    References:
        Vallado: 2022, p. 174, Algorithm 12

    Args:
        r (array_like): ECEF position vector in km

    Returns:
        tuple: (latgc, latgd, lon, hellp)
            latgc (float): Geocentric latitude in radians (-pi to pi)
            latgd (float): Geodetic latitude in radians (-pi to pi)
            lon (float): Longitude in radians (-2pi to 2pi)
            hellp (float): Height above the ellipsoid in km
    """
    # Compute magnitude of the position vector
    magr = np.linalg.norm(r)

    # Compute longitude (right ascension approximation)
    temp = np.sqrt(r[0] ** 2 + r[1] ** 2)
    if np.abs(temp) < const.SMALL:
        lon = np.sign(r[2]) * np.pi * 0.5
    else:
        lon = np.arctan2(r[1], r[0])

    # Adjust longitude to be within [-2π, 2π] range
    if np.abs(lon) >= np.pi:
        lon += const.TWOPI if lon < 0 else -const.TWOPI

    # Compute geodetic latitude as an initial approximation (declination)
    latgd = np.arcsin(r[2] / magr)

    # Iterate to refine geodetic latitude
    i, c = 0, 0
    olddelta = latgd + 10

    while np.abs(olddelta - latgd) >= const.SMALL and i < 10:
        olddelta = latgd
        sintemp = np.sin(latgd)
        c = const.RE / np.sqrt(1 - const.ECCEARTHSQRD * sintemp**2)
        latgd = np.arctan((r[2] + c * const.ECCEARTHSQRD * sintemp) / temp)
        i += 1

    # Calculate height above the ellipsoid
    if np.pi * 0.5 - np.abs(latgd) < np.radians(1):  # within 1 deg of poles
        hellp = (temp / np.cos(latgd)) - c
    else:
        s = c * (1 - const.ECCEARTHSQRD)
        hellp = r[2] / np.sin(latgd) - s

    # Compute geocentric latitude
    latgc = np.arcsin(r[2] / magr)

    return latgc, latgd, lon, hellp


def ecef2llb(r: ArrayLike) -> Tuple[float, float, float, float]:
    """Converts an ECEF position vector into geodetic and geocentric latitude,
    longitude, and height above the ellipsoid using the Borkowski method.

    References:
        Vallado: 2022, p. 175-176, Algorithm 13

    Args:
        r (array_like): ECEF position vector in km

    Returns:
        tuple: (latgc, latgd, lon, hellp)
            latgc (float): Geocentric latitude in radians (-pi to pi)
            latgd (float): Geodetic latitude in radians (-pi to pi)
            lon (float): Longitude in radians (-2pi to 2pi)
            hellp (float): Height above the ellipsoid in km
    """
    # Constants
    third = 1 / 3

    # Compute magnitude of the position vector
    magr = np.linalg.norm(r)

    # Earth semi-major and semi-minor axes
    a = const.RE
    b = np.sign(r[2]) * a * np.sqrt(1.0 - const.ECCEARTHSQRD)

    # Compute longitude (right ascension approximation)
    temp = np.sqrt(r[0] ** 2 + r[1] ** 2)
    if np.abs(temp) < const.SMALL:
        lon = np.sign(r[2]) * np.pi * 0.5
    else:
        lon = np.arctan2(r[1], r[0])

    # Adjust longitude to be within [-2π, 2π] range
    if np.abs(lon) >= np.pi:
        lon += const.TWOPI if lon < 0 else -const.TWOPI

    # Intermediate variables for polynomial solution
    atemp = 1 / (a * temp)
    e = (b * r[2] - a**2 + b**2) * atemp
    f = (b * r[2] + a**2 - b**2) * atemp
    p = 4 * third * (e * f + 1)
    q = 2 * (e**2 - f**2)
    d = p**3 + q**2

    # Solve the cubic equation based on the discriminant `d`
    if d > 0:
        nu = (np.sqrt(d) - q) ** third - (np.sqrt(d) + q) ** third
    else:
        sqrtp = np.sqrt(-p)
        nu = 2 * sqrtp * np.cos(third * np.arccos(q / (p * sqrtp)))

    # Intermediate variables for latitude and height computations
    g = 0.5 * (np.sqrt(e**2 + nu) + e)
    t = np.sqrt(g**2 + (f - nu * g) / (2 * g - e)) - g

    # Compute geodetic latitude and height above the ellipsoid
    latgd = np.arctan(a * (1 - t**2) / (2 * b * t))
    hellp = (temp - a * t) * np.cos(latgd) + (r[2] - b) * np.sin(latgd)

    # Compute geocentric latitude
    latgc = np.arcsin(r[2] / magr)

    return latgc, latgd, lon, hellp


########################################################################################
# Modified Equidistant Cylindrical (EQCM)
########################################################################################


def _compute_oes(r_rtn: np.ndarray, v_rtn: np.ndarray, ecc_tol: float):
    """Compute orbital elements"""
    # Calculate eccentricity vector and magnitude
    h_vec = np.cross(r_rtn, v_rtn)
    p = np.dot(h_vec, h_vec) / const.MU
    ecc_vec = np.cross(v_rtn, h_vec) / const.MU - r_rtn / np.linalg.norm(r_rtn)
    ecc = np.linalg.norm(ecc_vec)

    # Calculate semimajor axis and perigee unit vector
    a = p / (1 - ecc**2)
    perigee_unit = ecc_vec / ecc if ecc > ecc_tol else r_rtn / np.linalg.norm(r_rtn)
    lambda_perigee = np.arctan2(perigee_unit[1], perigee_unit[0])

    return p, ecc, a, perigee_unit, lambda_perigee


def _compute_rv_target(ecc, a, nu2, perigee_unit):
    """Compute the future position and velocity of the target in RTN2"""
    # Computes the future position and velocity of the target
    r2_tgt = a * (1 - ecc**2) / (1 + ecc * np.cos(nu2))
    p_vec = perigee_unit
    q_vec = np.cross([0, 0, 1], p_vec)

    r2_vec_tgt = r2_tgt * (np.cos(nu2) * p_vec + np.sin(nu2) * q_vec)
    v2_vec_tgt = np.sqrt(const.MU / (a * (1 - ecc**2))) * (
        -np.sin(nu2) * p_vec + (ecc + np.cos(nu2)) * q_vec
    )

    # Convert to RTN2 frame
    *_, rot_rtn1_to_rtn2 = rv2rsw(r2_vec_tgt, v2_vec_tgt)
    r_tgt_rtn2 = rot_rtn1_to_rtn2 @ r2_vec_tgt
    v_tgt_rtn2 = rot_rtn1_to_rtn2 @ v2_vec_tgt

    return r_tgt_rtn2, v_tgt_rtn2, r2_tgt


def _get_sez_rotation(phi, lambda_):
    """Computes the transformation matrix from RTN to SEZ frame."""
    sin_phi, cos_phi = np.sin(phi), np.cos(phi)
    sin_lambda, cos_lambda = np.sin(lambda_), np.cos(lambda_)

    return np.array(
        [
            [sin_phi * cos_lambda, sin_phi * sin_lambda, -cos_phi],
            [-sin_lambda, cos_lambda, 0],
            [cos_phi * cos_lambda, cos_phi * sin_lambda, sin_phi],
        ]
    )


def eci_to_eqcm_rtn(
    r_tgt_eci: ArrayLike,
    v_tgt_eci: ArrayLike,
    r_int_eci: ArrayLike,
    v_int_eci: ArrayLike,
    ecc_tol: float = 1e-7,
) -> tuple[np.ndarray, np.ndarray]:
    """Finds the relative position and velocity vectors in the Modified Equidistant
    Cylindrical (EQCM) frame given the ECI target and interceptor states with RTN (RSW)
    ordered components.

    References:
        Alfano: 2012

    Args:
        r_tgt_eci (array_like): ECI position vector of the target in km
        v_tgt_eci (array_like): ECI velocity vector of the target in km/s
        r_int_eci (array_like): ECI position vector of the interceptor in km
        v_int_eci (array_like): ECI velocity vector of the interceptor in km/s
        ecc_tol (float): Tolerance for eccentricity (defaults to 1e-7)

    Returns:
        tuple: (r_int_eqcm, v_int_eqcm)
            r_int_eqcm (np.ndarray): EQCM position vector of the interceptor in km
            v_int_eqcm (np.ndarray): EQCM velocity vector of the interceptor in km/s
    """
    # Compute rotation matrix from ECI to RTN1 frame for target
    r_tgt_rtn1, v_tgt_rtn1, rot_eci_to_rtn1 = rv2rsw(r_tgt_eci, v_tgt_eci)
    r_int_rtn1, v_int_rtn1 = rot_eci_to_rtn1 @ r_int_eci, rot_eci_to_rtn1 @ v_int_eci

    # Compute magnitudes
    mag_r_tgt, mag_r_int = np.linalg.norm(r_tgt_rtn1), np.linalg.norm(r_int_rtn1)

    # Compute lambda and phi rotation angles to go from target to interceptor
    # (lambda_tgt will be 0)
    sin_phi = r_int_rtn1[2] / mag_r_int
    phi = np.arcsin(sin_phi)
    lambda_ = np.arctan2(r_int_rtn1[1], r_int_rtn1[0])

    # Orbital elements of the target at present (nu1) and future (nu2) locations
    p_tgt, ecc_tgt, a_tgt, perigee_unit, lambda_perigee = _compute_oes(
        r_tgt_rtn1, v_tgt_rtn1, ecc_tol
    )
    nu1 = -lambda_perigee
    nu2 = lambda_ - lambda_perigee

    # Future position and velocity of target
    r_tgt_rtn2, v_tgt_rtn2, r2_tgt = _compute_rv_target(
        ecc_tgt, a_tgt, nu2, perigee_unit
    )

    # Convert interceptor components to SEZ frame
    rot_to_sez = _get_sez_rotation(phi, lambda_)
    r_int_sez = rot_to_sez @ r_int_rtn1
    v_int_sez = rot_to_sez @ v_int_rtn1

    # Position components in EQCM
    r_int_eqcm = np.zeros((3, 1))
    r_int_eqcm[0] = r_int_sez[2] - r_tgt_rtn2[0]

    ea0, _ = newtonnu(ecc_tgt, nu1)
    ea1, _ = newtonnu(ecc_tgt, nu2)

    # Fix quadrants for special cases
    if abs(ea1 - ea0) > np.pi:
        ea0 = ea0 + const.TWOPI if ea0 < 0 else ea0 - const.TWOPI

    # Calculate elliptic integrals of the second kind
    e0 = ellipeinc(ea0, ecc_tgt**2)
    e1 = ellipeinc(ea1, ecc_tgt**2)

    fixit = np.pi if e1 - e0 < 0 else 0
    r_int_eqcm[1] = a_tgt * (e1 - e0 + fixit)  # arclength value
    r_int_eqcm[2] = phi * r2_tgt

    # Velocity components in EQCM
    v_int_eqcm = np.zeros((3, 1))
    lambda_dot = v_int_sez[1] / (mag_r_int * np.cos(phi))
    v_int_eqcm[0] = v_int_sez[2] - v_tgt_rtn2[0]
    v_int_eqcm[1] = lambda_dot * r2_tgt - (v_tgt_rtn1[1] / mag_r_tgt) * mag_r_tgt
    v_int_eqcm[2] = (-v_int_sez[0] / mag_r_int) * r2_tgt

    return r_int_eqcm.ravel(), v_int_eqcm.ravel()


def eqcm_to_eci_rtn(
    r_tgt_eci: ArrayLike,
    v_tgt_eci: ArrayLike,
    r_int_eqcm: ArrayLike,
    v_int_eqcm: ArrayLike,
    ecc_tol: float = 1e-7,
) -> Tuple[np.ndarray, np.ndarray]:
    """Finds the interceptor position and velocity vectors in the ECI (RTN) frame given
    the target ECI state and interceptor Modified Equidistant Cylindrical (EQCM) state.

    References:
        Alfano: 2012

    Args:
        r_tgt_eci (array_like): ECI position vector of the target in km
        v_tgt_eci (array_like): ECI velocity vector of the target in km/s
        r_int_eqcm (array_like): EQCM position vector of the interceptor in km
        v_int_eqcm (array_like): EQCM velocity vector of the interceptor in km/s
        ecc_tol (float): Tolerance for eccentricity (defaults to 1e-7)

    Returns:
        tuple: (r_int_eci, v_int_eci)
            r_int_eci (np.ndarray): ECI position vector of the interceptor in km
            v_int_eci (np.ndarray): ECI velocity vector of the interceptor in km/s
    """
    # Convert target vectors to RTN1 frame
    r_tgt_rtn1, v_tgt_rtn1, rot_eci_to_rtn1 = rv2rsw(r_tgt_eci, v_tgt_eci)
    mag_r_tgt = np.linalg.norm(r_tgt_rtn1)

    # Compute orbital elements of the target at nu1
    p_tgt, ecc_tgt, a_tgt, perigee_unit, lambda_perigee = _compute_oes(
        r_tgt_rtn1, v_tgt_rtn1, ecc_tol
    )
    nu1 = -lambda_perigee

    # Compute nu2 and lambda from orbit arc
    arclength = r_int_eqcm[1]
    ea1, _ = newtonnu(ecc_tgt, nu1)

    # Tolerance and iteration definitions
    arclength_tol = 1e-3  # 1 m tolerance
    n_ea_its = 10

    # Compute eccentric anomaly correction
    if abs(float(arclength)) > arclength_tol:
        delta_ea = utils.inverse_elliptic2(arclength / a_tgt, ecc_tgt**2)
        e1 = ellipeinc(ea1, ecc_tgt**2)
        ea2 = ea1 + delta_ea

        # Iteratively refine eccentric anomaly correction
        for _ in range(n_ea_its):
            e2 = ellipeinc(ea2, ecc_tgt**2)
            arclength1 = a_tgt * (e2 - e1)
            if abs(arclength1 - arclength) < arclength_tol:
                break
            correction = arclength / (ea2 - ea1)
            ea2 -= (arclength1 - arclength) / correction

        _, nu2 = newtone(ecc_tgt, ea2)
    else:
        nu2 = nu1

    lambda_ = nu2 - nu1
    sin_lambda, cos_lambda = np.sin(lambda_), np.cos(lambda_)

    # Compute future target position and velocity at nu2
    r_tgt_rtn2, v_tgt_rtn2, r2_tgt = _compute_rv_target(
        ecc_tgt, a_tgt, nu2, perigee_unit
    )

    # Compute phi and interceptor position unit vector in RTN1
    phi = r_int_eqcm[2] / r2_tgt
    r_int_unit_rtn1 = np.array(
        [cos_lambda * np.cos(phi), sin_lambda * np.cos(phi), np.sin(phi)]
    )

    # Convert to SEZ frame
    rot_to_sez = _get_sez_rotation(phi, lambda_)
    r_int_unit_sez = rot_to_sez @ r_int_unit_rtn1

    # Compute proper scaling from Z component of interceptor
    r_int_sez = np.zeros(3)
    r_int_sez[2] = r_int_eqcm[0] + r_tgt_rtn2[0]
    mag_r_int = r_int_sez[2] / r_int_unit_sez[2]
    r_int_rtn1 = mag_r_int * r_int_unit_rtn1

    # Compute velocity components in RTN1
    lambda_dot = (v_int_eqcm[1] + (v_tgt_rtn1[1] / mag_r_tgt) * mag_r_tgt) / r2_tgt

    v_int_sez = np.array(
        [
            (-v_int_eqcm[2] / r2_tgt) * mag_r_int,
            lambda_dot * mag_r_int * np.cos(phi),
            v_int_eqcm[0] + v_tgt_rtn2[0],
        ]
    )

    v_int_rtn1 = rot_to_sez.T @ v_int_sez

    # Convert all back to original ECI frame
    r_int_eci = rot_eci_to_rtn1.T @ r_int_rtn1
    v_int_eci = rot_eci_to_rtn1.T @ v_int_rtn1

    return r_int_eci, v_int_eci


########################################################################################
# Miscellaneous
########################################################################################


def perifocal_transform(
    i: float, raan: float, w: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transform classical elements from equatorial (ECI) frame to perifocal frame.

    References:
        Vallado: 2022, p. 171, Equation 3-31

    Args:
        i (float): Inclination in radians
        raan (float): Right ascension of the ascending node in radians
        w (float): Argument of periapsis in radians

    Returns:
        tuple: (p_, q_, w_)
            p_ (np.ndarray): P vector in perifocal frame
            q_ (np.ndarray): Q vector in perifocal frame
            w_ (np.ndarray): W vector in perifocal frame
    """
    p_ = np.array(
        [
            np.cos(w) * np.cos(raan) - np.sin(w) * np.sin(raan) * np.cos(i),
            np.cos(w) * np.sin(raan) + np.sin(w) * np.cos(raan) * np.cos(i),
            np.sin(w) * np.sin(i),
        ]
    )

    q_ = np.array(
        [
            -np.sin(w) * np.cos(raan) - np.cos(w) * np.sin(raan) * np.cos(i),
            -np.sin(w) * np.sin(raan) + np.cos(w) * np.cos(raan) * np.cos(i),
            np.cos(w) * np.sin(i),
        ]
    )

    w_ = np.array([np.sin(raan) * np.sin(i), -np.cos(raan) * np.sin(i), np.cos(i)])

    return p_, q_, w_



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 21 June 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from enum import Enum
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ...time.data import IAU80Array
from .. import frame_conversions as fc
from ..newton import newtonnu
from ....constants import KM2M, MUM, SMALL, TWOPI
from ....mathtime.vector import unit


class AnomalyType(Enum):
    MEAN_A = "meana"
    TRUE_A = "truea"
    MEAN_N = "meann"
    TRUE_N = "truen"


def is_mean_anomaly(anom_type: AnomalyType) -> bool:
    """Checks if the anomaly type is a mean anomaly.

    Args:
        anom_type (AnomalyType): Anomaly type

    Returns:
        bool: True if the anomaly type is a mean anomaly
    """
    return anom_type in [AnomalyType.MEAN_A, AnomalyType.MEAN_N]


########################################################################################
# Classical <-> Cartesian Elements
# TODO: This section could use some cleanup
########################################################################################


def _compute_partials_a(a, n, reci_m, veci_m):
    """Compute partial derivatives of a w.r.t. (rx, ry, rz, vx, vy, vz)."""
    tm_a = np.zeros(6)
    p0 = 2.0 * a**2 / np.linalg.norm(reci_m) ** 3
    p1 = 2.0 / (n**2 * a)
    tm_a[:3] = p0 * reci_m
    tm_a[3:] = p1 * veci_m
    return tm_a


def _compute_partials_ecc(reci_m, veci_m, ecc_vec, ecc):
    """Compute partial derivatives of ecc w.r.t. (rx, ry, rz, vx, vy, vz)."""
    rx, ry, rz = reci_m
    vx, vy, vz = veci_m
    tm_ecc = np.zeros(6)
    p0 = 1.0 / (MUM * ecc)

    magr = np.linalg.norm(reci_m)
    magr3 = magr**3

    tm_ecc[0] = -p0 * (
        ((vx * vy - MUM * rx * ry / magr3) * ecc_vec[1])
        + ((vx * vz - MUM * rx * rz / magr3) * ecc_vec[2])
        - (vy**2 + vz**2 - MUM / magr + MUM * rx**2 / magr3) * ecc_vec[0]
    )
    tm_ecc[1] = -p0 * (
        ((vx * vy - MUM * rx * ry / magr3) * ecc_vec[0])  # TODO: check
        + ((vy * vz - MUM * ry * rz / magr3) * ecc_vec[2])
        - (vx**2 + vz**2 - MUM / magr + MUM * ry**2 / magr3) * ecc_vec[1]
    )
    tm_ecc[2] = -p0 * (
        ((vx * vz - MUM * rx * rz / magr3) * ecc_vec[0])
        + ((vy * vz - MUM * ry * rz / magr3) * ecc_vec[1])
        - (vx**2 + vy**2 - MUM / magr + MUM * rz**2 / magr3) * ecc_vec[2]
    )
    tm_ecc[3] = -p0 * (
        ((rx * vy - 2 * ry * vx) * ecc_vec[1])
        + ((ry * vy + rz * vz) * ecc_vec[0])
        + ((rx * vz - 2 * rz * vx) * ecc_vec[2])
    )
    tm_ecc[4] = -p0 * (
        ((ry * vx - 2 * rx * vy) * ecc_vec[0])
        + ((rx * vx + rz * vz) * ecc_vec[1])
        + ((ry * vz - 2 * rz * vy) * ecc_vec[2])
    )
    tm_ecc[5] = -p0 * (
        ((rx * vx + ry * vy) * ecc_vec[2])
        + ((rz * vx - 2 * rx * vz) * ecc_vec[0])
        + ((rz * vy - 2 * ry * vz) * ecc_vec[1])
    )

    return tm_ecc


def _compute_partials_incl(reci_m, veci_m, h_vec, node):
    """Compute partial derivatives of inclination w.r.t. (rx, ry, rz, vx, vy, vz)."""
    rx, ry, rz = reci_m
    vx, vy, vz = veci_m
    h = np.linalg.norm(h_vec)
    tm_incl = np.zeros(6)
    p3 = 1.0 / node
    tm_incl[0] = -p3 * (vy - h_vec[2] * (vy * h_vec[2] - vz * h_vec[1]) / h**2)
    tm_incl[1] = p3 * (vx - h_vec[2] * (vx * h_vec[2] - vz * h_vec[0]) / h**2)
    tm_incl[2] = -p3 * (h_vec[2] * (vy * h_vec[0] - vx * h_vec[1]) / h**2)
    tm_incl[3] = p3 * (ry - h_vec[2] * (ry * h_vec[2] - rz * h_vec[1]) / h**2)
    tm_incl[4] = -p3 * (rx - h_vec[2] * (rx * h_vec[2] - rz * h_vec[0]) / h**2)
    tm_incl[5] = p3 * (h_vec[2] * (ry * h_vec[0] - rx * h_vec[1]) / h**2)

    return tm_incl


def _compute_partials_node(reci_m, veci_m, node_vec, node):
    """Compute partial derivatives of node w.r.t. (rx, ry, rz, vx, vy, vz)."""
    rx, ry, rz = reci_m
    vx, vy, vz = veci_m
    tm_node = np.zeros(6)
    p4 = 1.0 / (node**2)
    tm_node[0] = -p4 * vz * node_vec[1]
    tm_node[1] = p4 * vz * node_vec[0]
    tm_node[2] = p4 * (vx * node_vec[1] - vy * node_vec[0])
    tm_node[3] = p4 * rz * node_vec[1]
    tm_node[4] = -p4 * rz * node_vec[0]
    tm_node[5] = p4 * (ry * node_vec[0] - rx * node_vec[1])

    return tm_node


def _compute_partials_argp(
    tm_ecc, reci_m, veci_m, ecc_vec, ecc, h_vec, node, n_dot_e, w_scale
):
    """Compute partial derivatives of argp w.r.t. (rx, ry, rz, vx, vy, vz)."""
    rx, ry, rz = reci_m
    vx, vy, vz = veci_m
    magr = np.linalg.norm(reci_m)
    tm_argp = np.zeros(6)
    tm_argp[0] = (
        -h_vec[1] * (vy**2 + vz**2 - MUM / magr + MUM * rx**2 / magr**3)
        - h_vec[0] * (vx * vy - MUM * rx * ry / magr**3)
        + vz * MUM * ecc_vec[0]
    )
    tm_argp[0] = (
        tm_argp[0] / (MUM * node * ecc)
        + vz * h_vec[1] * n_dot_e / (node**3 * ecc)
        - tm_ecc[0] * n_dot_e / (node * ecc**2)
    ) * w_scale

    tm_argp[1] = (
        h_vec[0] * (vx**2 + vz**2 - MUM / magr + MUM * ry**2 / magr**3)
        + h_vec[1] * (vx * vy - MUM * rx * ry / magr**3)
        + vz * MUM * ecc_vec[1]
    )
    tm_argp[1] = (
        tm_argp[1] / (MUM * node * ecc)
        - vz * h_vec[0] * n_dot_e / (node**3 * ecc)
        - tm_ecc[1] * n_dot_e / (node * ecc**2)
    ) * w_scale

    tm_argp[2] = (
        -h_vec[1] * (vx * vz - MUM * rx * rz / magr**3)
        + h_vec[0] * (vy * vz - MUM * ry * rz / magr**3)
        + vx * MUM * ecc_vec[0]
        + vy * MUM * ecc_vec[1]
    )
    tm_argp[2] = (
        -tm_argp[2] / (MUM * node * ecc)
        + (vy * h_vec[0] - vx * h_vec[1]) * n_dot_e / (node**3 * ecc)
        - tm_ecc[2] * n_dot_e / (node * ecc**2)
    ) * w_scale

    tm_argp[3] = (
        (rx * vy - 2 * ry * vx) * h_vec[0]
        - h_vec[1] * (ry * vy + rz * vz)
        + rz * MUM * ecc_vec[0]
    )
    tm_argp[3] = (
        -tm_argp[3] / (MUM * node * ecc)
        - rz * h_vec[1] * n_dot_e / (node**3 * ecc)
        - tm_ecc[3] * n_dot_e / (node * ecc**2)
    ) * w_scale

    tm_argp[4] = (
        -(ry * vx - 2 * rx * vy) * h_vec[1]
        + h_vec[0] * (rx * vx + rz * vz)
        + rz * MUM * ecc_vec[1]
    )
    tm_argp[4] = (
        -tm_argp[4] / (MUM * node * ecc)
        + rz * h_vec[0] * n_dot_e / (node**3 * ecc)
        - tm_ecc[4] * n_dot_e / (node * ecc**2)
    ) * w_scale

    tm_argp[5] = (
        -(rz * vx - 2 * rx * vz) * h_vec[1]
        + h_vec[0] * (rz * vy - 2 * ry * vz)
        - rx * MUM * ecc_vec[0]
        - ry * MUM * ecc_vec[1]
    )
    tm_argp[5] = (
        -tm_argp[5] / (MUM * node * ecc)
        + (rx * h_vec[1] - ry * h_vec[0]) * n_dot_e / (node**3 * ecc)
        - tm_ecc[5] * n_dot_e / (node * ecc**2)
    ) * w_scale

    return tm_argp


def _compute_partials_nu(
    tm_ecc, reci_m, veci_m, ecc, ecc_term, r_dot_v, r_dot_e, nu_scale
):
    """Compute partial derivatives of nu/M w.r.t. (rx, ry, rz, vx, vy, vz)."""
    rx, ry, rz = reci_m
    vx, vy, vz = veci_m
    magr = np.linalg.norm(reci_m)
    tm_nu = np.zeros(6)
    tm_nu[0] = (
        ry * (vx * vy - MUM * rx * ry / magr**3)
        - rx * ecc_term
        + rz * (vx * vz - MUM * rx * rz / magr**3)
        - rx * (vy**2 + vz**2 - MUM / magr + MUM * rx**2 / magr**3)
        + vx * r_dot_v
    )
    tm_nu[0] = (
        -tm_nu[0] / (MUM * magr * ecc)
        - rx * r_dot_e / (magr**3 * ecc)
        - tm_ecc[0] * r_dot_e / (magr * ecc**2)
    ) * nu_scale

    tm_nu[1] = (
        rx * (vx * vy - MUM * rx * ry / magr**3)
        - ry * ecc_term
        + rz * (vy * vz - MUM * ry * rz / magr**3)
        - ry * (vx**2 + vz**2 - MUM / magr + MUM * ry**2 / magr**3)
        + vy * r_dot_v
    )
    tm_nu[1] = (
        -tm_nu[1] / (MUM * magr * ecc)
        - ry * r_dot_e / (magr**3 * ecc)
        - tm_ecc[1] * r_dot_e / (magr * ecc**2)
    ) * nu_scale

    tm_nu[2] = (
        rx * (vx * vz - MUM * rx * rz / magr**3)
        - rz * ecc_term
        + ry * (vy * vz - MUM * ry * rz / magr**3)
        - rz * (vx**2 + vy**2 - MUM / magr + MUM * rz**2 / magr**3)
        + vz * r_dot_v
    )
    tm_nu[2] = (
        -tm_nu[2] / (MUM * magr * ecc)
        - rz * r_dot_e / (magr**3 * ecc)
        - tm_ecc[2] * r_dot_e / (magr * ecc**2)
    ) * nu_scale

    tm_nu[3] = (
        ry * (rx * vy - 2 * ry * vx)
        + rx * (ry * vy + rz * vz)
        + rz * (rx * vz - 2 * rz * vx)
    )
    tm_nu[3] = (
        -tm_nu[3] / (MUM * magr * ecc) - tm_ecc[3] * r_dot_e / (magr * ecc**2)
    ) * nu_scale

    tm_nu[4] = (
        rx * (ry * vx - 2 * rx * vy)
        + ry * (rx * vx + rz * vz)
        + rz * (ry * vz - 2 * rz * vy)
    )
    tm_nu[4] = (
        -tm_nu[4] / (MUM * magr * ecc) - tm_ecc[4] * r_dot_e / (magr * ecc**2)
    ) * nu_scale

    tm_nu[5] = (
        rz * (rx * vx + ry * vy)
        + rx * (rz * vx - 2 * rx * vz)
        + ry * (rz * vy - 2 * ry * vz)
    )
    tm_nu[5] = (
        -tm_nu[5] / (MUM * magr * ecc) - tm_ecc[5] * r_dot_e / (magr * ecc**2)
    ) * nu_scale

    return tm_nu


def covct2cl(
    cartcov: ArrayLike, cartstate: ArrayLike, use_mean_anom: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from Cartesian elements to classical orbital
    elements.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        cartcov (array_like): 6x6 Cartesian covariance matrix in m and m/s
        cartstate (array_like): 6x1 Cartesian orbit state in km and km/s
                                (rx, ry, rz, vx, vy, vz)
        use_mean_anom (bool): Flag to use mean anomaly instead of true anomaly
                              (defaults to False)

    Returns:
        tuple: (classcov, tm)
            classcov (np.ndarray): 6x6 Classical orbital elements covariance matrix
                                   in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """
    # Parse the input state vector
    reci_m = np.array(cartstate[:3]) * KM2M
    veci_m = np.array(cartstate[3:]) * KM2M

    # Convert to classical orbital elements
    _, a, ecc, incl, omega, argp, nu, *_ = fc.rv2coe(reci_m / KM2M, veci_m / KM2M)
    a *= KM2M
    n = np.sqrt(MUM / a**3)

    # Common quantities
    sqrt1me2 = np.sqrt(1 - ecc**2)
    magr = np.linalg.norm(reci_m)
    magv = np.linalg.norm(veci_m)

    # Eccentricity vector
    r_dot_v = np.dot(reci_m, veci_m)
    ecc_term = magv**2 - MUM / magr
    ecc_vec = (ecc_term * reci_m - r_dot_v * veci_m) / MUM

    # Node vector
    h_vec = np.cross(reci_m, veci_m)
    node_vec = np.cross([0, 0, 1], h_vec)
    node = np.linalg.norm(node_vec)

    # Additional terms for argument of periapsis and true anomaly
    n_dot_e = np.dot(node_vec, ecc_vec)
    sign_w = np.sign((magv**2 - MUM / magr) * reci_m[2] - r_dot_v * veci_m[2])
    cos_w = n_dot_e / (ecc * node)
    w_scale = -sign_w / np.sqrt(1 - cos_w**2)

    # Additional terms for true anomaly
    r_dot_e = np.dot(reci_m, ecc_vec)
    cos_nu = r_dot_e / (magr * ecc)
    sign_nu = np.sign(r_dot_v)
    nu_scale = -sign_nu / np.sqrt(1 - cos_nu**2)

    # Compute partial derivatives
    tm = np.zeros((6, 6))

    # Compute partials
    tm[0, :] = _compute_partials_a(a, n, reci_m, veci_m)
    tm[1, :] = _compute_partials_ecc(reci_m, veci_m, ecc_vec, ecc)
    tm[2, :] = _compute_partials_incl(reci_m, veci_m, h_vec, node)
    tm[3, :] = _compute_partials_node(reci_m, veci_m, node_vec, node)
    tm[4, :] = _compute_partials_argp(
        tm[1, :], reci_m, veci_m, ecc_vec, ecc, h_vec, node, n_dot_e, w_scale
    )
    tm[5, :] = _compute_partials_nu(
        tm[1, :], reci_m, veci_m, ecc, ecc_term, r_dot_v, r_dot_e, nu_scale
    )

    # Update partials for mean anomaly, if specified
    if use_mean_anom:
        dmdnu = (sqrt1me2**2) ** 1.5 / (1 + ecc * np.cos(nu)) ** 2
        dmde = (
            -np.sin(nu)
            * (
                (ecc * np.cos(nu) + 1)
                * (ecc + np.cos(nu))
                / np.sqrt((ecc + np.cos(nu)) ** 2)
                + 1
                - 2 * ecc**2
                - ecc**3 * np.cos(nu)
            )
            / ((ecc * np.cos(nu) + 1) ** 2 * sqrt1me2)
        )
        tm[5, :] = tm[5, :] * dmdnu + tm[1, :] * dmde

    # Calculate the classical covariance matrix
    classcov = tm @ cartcov @ tm.T

    return classcov, tm


def covcl2ct(
    classcov: ArrayLike, classstate: ArrayLike, use_mean_anom: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from classical elements to Cartesian elements.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        classcov (array_like): 6x6 Classical orbital elements covariance matrix
                               in m and m/s
        classstate (array_like): 6x1 Classical orbital elements in km and radians
                                (a, ecc, incl, node, argp, nu or m)
        use_mean_anom (bool): Flag to use mean anomaly instead of true anomaly
                              (defaults to False)

    Returns:
        tuple: (cartcov, tm)
            cartcov (np.ndarray): 6x6 Cartesian covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """

    def update_row_with_mean_anomaly(mat, row):
        if use_mean_anom:
            mat[row, 5] /= dmdnu
            mat[row, 1] -= mat[row, 5] * dmde
        return mat

    # Parse the classical elements
    a, ecc, incl, raan, argp, anom = classstate
    a *= KM2M

    # Convert anomaly as needed
    if use_mean_anom:
        e, nu = fc.newtonm(ecc, anom)
    else:
        nu = anom
        e, _ = fc.newtonnu(ecc, nu)

    # Compute trigonometric values
    sin_inc, cos_inc = np.sin(incl), np.cos(incl)
    sin_raan, cos_raan = np.sin(raan), np.cos(raan)
    sin_w, cos_w = np.sin(argp), np.cos(argp)
    sin_nu, cos_nu = np.sin(nu), np.cos(nu)

    # Define PQW to ECI transformation elements (p. 168)
    p11 = cos_raan * cos_w - sin_raan * sin_w * cos_inc
    p12 = -cos_raan * sin_w - sin_raan * cos_w * cos_inc
    p13 = sin_raan * sin_inc
    p21 = sin_raan * cos_w + cos_raan * sin_w * cos_inc
    p22 = -sin_raan * sin_w + cos_raan * cos_w * cos_inc
    p23 = -cos_raan * sin_inc
    p31 = sin_w * sin_inc
    p32 = cos_w * sin_inc

    # Define constants for efficiency
    p0 = np.sqrt(MUM / (a * (1 - ecc**2)))
    p1 = (1 - ecc**2) / (1 + ecc * cos_nu)
    p2 = 1 / (2 * a) * p0
    p3 = (2 * a * ecc + a * cos_nu + a * cos_nu * ecc**2) / ((1 + ecc * cos_nu) ** 2)
    p4 = ecc * MUM / (a * (1 - ecc**2) ** 2 * p0)
    p5 = a * p1
    p6 = a * (1 - ecc**2) / ((1 + ecc * cos_nu) ** 2)

    dmdnu = (1 - ecc**2) ** 1.5 / (1 + ecc * cos_nu) ** 2
    dmde = (
        -sin_nu
        * (
            (ecc * cos_nu + 1) * (ecc + cos_nu) / np.sqrt((ecc + cos_nu) ** 2)
            + 1
            - 2 * ecc**2
            - ecc**3 * cos_nu
        )
        / ((ecc * cos_nu + 1) ** 2 * np.sqrt(1 - ecc**2))
    )

    # Compute partial derivatives
    tm = np.zeros((6, 6))

    # Partials of (a, ecc, incl, node, argp, nu) w.r.t. rx
    tm[0, 0] = p1 * (p11 * cos_nu + p12 * sin_nu)
    tm[0, 1] = -p3 * (p11 * cos_nu + p12 * sin_nu)
    tm[0, 2] = p5 * p13 * (sin_w * cos_nu + cos_w * sin_nu)
    tm[0, 3] = -p5 * (p21 * cos_nu + p22 * sin_nu)
    tm[0, 4] = p5 * (p12 * cos_nu - p11 * sin_nu)
    tm[0, 5] = p6 * (-p11 * sin_nu + p12 * (ecc + cos_nu))
    tm = update_row_with_mean_anomaly(tm, 0)

    # Partials of (a, ecc, incl, node, argp, nu) w.r.t. ry
    tm[1, 0] = p1 * (p21 * cos_nu + p22 * sin_nu)
    tm[1, 1] = -p3 * (p21 * cos_nu + p22 * sin_nu)
    tm[1, 2] = p5 * p23 * (sin_w * cos_nu + cos_w * sin_nu)
    tm[1, 3] = p5 * (p11 * cos_nu + p12 * sin_nu)
    tm[1, 4] = p5 * (p22 * cos_nu - p21 * sin_nu)
    tm[1, 5] = p6 * (-p21 * sin_nu + p22 * (ecc + cos_nu))
    tm = update_row_with_mean_anomaly(tm, 1)

    # Partials of (a, ecc, incl, node, argp, nu) w.r.t. rz
    tm[2, 0] = p1 * (p31 * cos_nu + p32 * sin_nu)
    tm[2, 1] = -p3 * sin_inc * (cos_w * sin_nu + sin_w * cos_nu)
    tm[2, 2] = p5 * cos_inc * (cos_w * sin_nu + sin_w * cos_nu)
    tm[2, 3] = 0.0
    tm[2, 4] = p5 * sin_inc * (cos_w * cos_nu - sin_w * sin_nu)
    tm[2, 5] = p6 * (-p31 * sin_nu + p32 * (ecc + cos_nu))
    tm = update_row_with_mean_anomaly(tm, 2)

    # Partials of (a, ecc, incl, node, argp, nu) w.r.t. vx
    tm[3, 0] = p2 * (p11 * sin_nu - p12 * (ecc + cos_nu))
    tm[3, 1] = -p4 * (p11 * sin_nu - p12 * (ecc + cos_nu)) + p12 * p0
    tm[3, 2] = -p0 * sin_raan * (p31 * sin_nu - p32 * (ecc + cos_nu))
    tm[3, 3] = p0 * (p21 * sin_nu - p22 * (ecc + cos_nu))
    tm[3, 4] = -p0 * (p12 * sin_nu + p11 * (ecc + cos_nu))
    tm[3, 5] = -p0 * (p11 * cos_nu + p12 * sin_nu)
    tm = update_row_with_mean_anomaly(tm, 3)

    # Partials of (a, ecc, incl, node, argp, nu) w.r.t. vy
    tm[4, 0] = p2 * (p21 * sin_nu - p22 * (ecc + cos_nu))
    tm[4, 1] = -p4 * (p21 * sin_nu - p22 * (ecc + cos_nu)) + p22 * p0
    tm[4, 2] = p0 * cos_raan * (p31 * sin_nu - p32 * (ecc + cos_nu))
    tm[4, 3] = p0 * (-p11 * sin_nu + p12 * (ecc + cos_nu))
    tm[4, 4] = -p0 * (p22 * sin_nu + p21 * (ecc + cos_nu))
    tm[4, 5] = -p0 * (p21 * cos_nu + p22 * sin_nu)
    tm = update_row_with_mean_anomaly(tm, 4)

    # Partials of (a, ecc, incl, node, argp, nu) w.r.t. vz
    tm[5, 0] = p2 * (p31 * sin_nu - p32 * (ecc + cos_nu))
    tm[5, 1] = -p4 * (p31 * sin_nu - p32 * (ecc + cos_nu)) + p32 * p0
    tm[5, 2] = p0 * cos_inc * (cos_w * cos_nu - sin_w * sin_nu + ecc * cos_w)
    tm[5, 3] = 0.0
    tm[5, 4] = -p0 * (p32 * sin_nu + p31 * (ecc + cos_nu))
    tm[5, 5] = -p0 * (p31 * cos_nu + p32 * sin_nu)
    tm = update_row_with_mean_anomaly(tm, 5)

    # Calculate the Cartesian covariance matrix
    cartcov = tm @ classcov @ tm.T

    return cartcov, tm


########################################################################################
# Equinoctial <-> Cartesian Elements
########################################################################################


def newton_mean_anomaly(
    meanlon_m: float, af: float, ag: float, tol: float = 1e-8, max_iter: int = 25
) -> float:
    """Solves for the equinoctial anomaly (F) using Newton's method.

    Args:
        meanlon_m (float): Mean longitude (mean anomaly + mean argument of latitude)
        af (float): Component of the eccentricity vector in equinoctial elements
        ag (float): Component of the eccentricity vector in equinoctial elements
        tol (float, optional): Convergence tolerance (defaults to 1e-8)
        max_iter (int, optional): Maximum number of iterations (default is 25)

    Returns:
        float: The solved equinoctial anomaly (F)
    """
    # Initial guess
    f0 = meanlon_m
    for ktr in range(max_iter):
        # Compute the function value and its derivative
        f1 = f0 - (f0 + ag * np.cos(f0) - af * np.sin(f0) - meanlon_m) / (
            1 - ag * np.sin(f0) - af * np.cos(f0)
        )

        # Check for convergence
        if abs(f1 - f0) < tol:
            return f1

        # Update the guess
        f0 = f1

    # If the loop completes without convergence, raise an error
    raise RuntimeError(f"Newton method failed to converge after {max_iter} iterations.")


def covct2eq(
    cartcov: ArrayLike, cartstate: ArrayLike, fr: int, use_mean_anom: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from Cartesian to equinoctial elements.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        cartcov (array_like): 6x6 Cartesian covariance matrix in m and m/s
        cartstate (array_like): 6x1 Cartesian state vector in km and km/s
                                (rx, ry, rz, vx, vy, vz)
        fr (int): Retrograde factor (+1 or -1)
        use_mean_anom (bool): Flag to use mean anomaly instead of true anomaly
                              (defaults to False)

    Returns:
        tuple: (eqcov, tm)
            eqcov (np.ndarray): 6x6 Equinoctial covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """
    # Parse input vectors
    reci_m = np.array(cartstate[:3]) * KM2M
    veci_m = np.array(cartstate[3:]) * KM2M
    magr = np.linalg.norm(reci_m)
    magv = np.linalg.norm(veci_m)

    # Classical orbital elements
    a = 1 / (2 / magr - magv**2 / MUM)
    n = np.sqrt(MUM / a**3)
    h_vec = np.cross(reci_m, veci_m)
    w_vec = h_vec / np.linalg.norm(h_vec)
    chi = w_vec[0] / (1 + fr * w_vec[2])
    psi = -w_vec[1] / (1 + fr * w_vec[2])

    # Equinoctial components
    p0 = 1 / (1 + chi**2 + psi**2)
    f_vec = p0 * np.array([1 - chi**2 + psi**2, 2 * chi * psi, -2 * fr * chi])
    g_vec = p0 * np.array([2 * fr * chi * psi, fr * (1 + chi**2 - psi**2), 2 * psi])

    # Compute eccentricity vector
    r_dot_v = np.dot(reci_m, veci_m)
    p1 = magv**2 - MUM / magr
    ecc_vec = (p1 * reci_m - r_dot_v * veci_m) / MUM

    # Get eccentricity components
    af = np.dot(ecc_vec, f_vec)
    ag = np.dot(ecc_vec, g_vec)

    # Intermediate terms
    x = np.dot(reci_m, f_vec)
    y = np.dot(reci_m, g_vec)

    b = 1 / (1 + np.sqrt(1 - af**2 - ag**2))
    p0 = 1 / (a * np.sqrt(1 - af**2 - ag**2))
    sinf = ag + p0 * ((1 - ag**2 * b) * y - ag * af * b * x)
    cosf = af + p0 * ((1 - af**2 * b) * x - ag * af * b * y)
    f = np.arctan2(sinf, cosf) % TWOPI

    xd = n * a**2 / magr * (af * ag * b * np.cos(f) - (1 - ag**2 * b) * np.sin(f))
    yd = n * a**2 / magr * ((1 - af**2 * b) * np.cos(f) - af * ag * b * np.sin(f))

    # Compute partial derivatives
    a_ = np.sqrt(MUM * a)
    b_ = np.sqrt(1 - ag**2 - af**2)
    c_ = 1 + chi**2 + psi**2

    a_term = a_ / magr**3
    term1_b = a / (1 + b_)
    term2_b = 1 / b_

    partxdaf = a * xd * yd / (a_ * b_) - a_term * (term1_b * ag * x + term2_b * x * y)
    partydaf = -a * xd**2 / (a_ * b_) - a_term * (term1_b * ag * y - term2_b * x**2)
    partxdag = a * yd**2 / (a_ * b_) + a_term * (term1_b * af * x - term2_b * y**2)
    partydag = -a * xd * yd / (a_ * b_) + a_term * (term1_b * af * y + term2_b * x * y)

    # Initialize transformation matrix
    tm = np.zeros((6, 6))

    # Partials of sma w.r.t. (rx ry rz vx vy vz)
    if use_mean_anom:
        p0 = 2 * a**2 / magr**3
        p1 = 2 / (n**2 * a)
    else:
        p0 = -3 * n * a / magr**3
        p1 = -3 / (n * a**2)

    tm[0, :3] = p0 * reci_m
    tm[0, 3:] = p1 * veci_m

    # Partials of v w.r.t. ag
    tm_v_ag_parts = partxdag * f_vec + partydag * g_vec

    # Partials of af w.r.t. (rx ry rz vx vy vz)
    p0 = 1 / MUM
    af_term = -a * b * af * b_ / (magr**3)
    chi_psi_term = ag * (chi * xd - psi * fr * yd) / (a_ * b_)
    tm[1, :3] = af_term * reci_m - chi_psi_term * w_vec + (b_ / a_) * tm_v_ag_parts

    xy_term = p0 * ((2 * x * yd - xd * y) * g_vec - y * yd * f_vec)
    psi_chi_term = ag * (psi * fr * y - chi * x) / (a_ * b_)
    tm[1, 3:] = xy_term - psi_chi_term * w_vec

    # Partials of v w.r.t. af
    tm_v_af_parts = partxdaf * f_vec + partydaf * g_vec

    # Partials of af w.r.t. (rx ry rz vx vy vz)
    ag_term = -a * b * ag * b_ / (magr**3)
    chi_psi_term_af = af * (chi * xd - psi * fr * yd) / (a_ * b_)
    tm[2, :3] = ag_term * reci_m + chi_psi_term_af * w_vec - (b_ / a_) * tm_v_af_parts

    xy_term = p0 * ((2 * xd * y - x * yd) * f_vec - x * xd * g_vec)
    psi_chi_term_af = af * (psi * fr * y - chi * x) / (a_ * b_)
    tm[2, 3:] = xy_term + psi_chi_term_af * w_vec

    # Partials of chi w.r.t. (rx ry rz vx vy vz)
    den = 2 * a_ * b_
    tm[3, :3] = -c_ * yd * w_vec / den
    tm[3, 3:] = c_ * y * w_vec / den

    # Partials of psi w.r.t. (rx ry rz vx vy vz)
    tm[4, :3] = -fr * c_ * xd * w_vec / den
    tm[4, 3:] = fr * c_ * x * w_vec / den

    # Partials of true/mean anomaly w.r.t. (rx ry rz vx vy vz)
    tm[5, :] = 0
    if use_mean_anom:
        tm[5, :3] = (
            -veci_m / a_
            + (chi * xd - psi * fr * yd) * w_vec / (a_ * b_)
            - (b * b_ / a_) * (ag * tm_v_ag_parts + af * tm_v_af_parts)
        )
        tm[5, 3:] = (
            -2.0 * reci_m / a_
            + (af * tm[2, 3:6] - ag * tm[1, 3:6]) / (1 + b_)
            + (fr * psi * y - chi * x) * w_vec / a_
        )

    # Compute the equinoctial covariance matrix
    eqcov = tm @ cartcov @ tm.T

    return eqcov, tm


def coveq2ct(
    eqcov: ArrayLike, eqstate: ArrayLike, fr: int, anom_type: AnomalyType
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from equinoctial to cartesian elements.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        eqcov (array_like): 6x6 equinoctial covariance matrix in m and m/s
        eqstate (array_like): 6x1 equinoctial orbit state in km and km/s
                              (a/n, af, ag, chi, psi, lm/ln)
        fr (int): Retrograde factor (+1 or -1)
        anom_type (AnomalyType): Anomaly type (MEAN_A, TRUE_A, MEAN_N, TRUE_N)

    Returns:
        tuple: (cartcov, tm)
            cartcov (np.ndarray): 6x6 Cartesian covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix

    TODO:
        - Return cartesian covariance matrix not as expected?
    """
    # Parse eqstate and anomaly-dependent calculations
    use_anom_a = anom_type in {AnomalyType.TRUE_A, AnomalyType.MEAN_A}
    if use_anom_a:
        a = eqstate[0] * KM2M  # in meters
        n = np.sqrt(MUM / a**3)
    else:
        n = eqstate[0]  # rad/s
        a = (MUM / n**2) ** (1 / 3)  # in meters

    # Get orbital elements
    af, ag, chi, psi = eqstate[1:5]
    use_mean_anom = anom_type in {AnomalyType.MEAN_A, AnomalyType.MEAN_N}
    meanlon_m = eqstate[5] if use_mean_anom else None
    meanlon_nu = eqstate[5] if not use_mean_anom else None
    omega = np.arctan2(chi, psi)
    argp = np.arctan2(ag, af) - fr * omega
    ecc = np.sqrt(af**2 + ag**2)

    # Update for true anomaly
    if not use_mean_anom:
        nu = np.mod(meanlon_nu - fr * omega - argp, TWOPI)
        _, m = newtonnu(ecc, nu)
        meanlon_m = np.mod(fr * omega + argp + m, TWOPI)

    # Convert equinoctial to cartesian state
    reci, veci = fc.eq2rv(a / KM2M, af, ag, chi, psi, meanlon_m, fr)
    reci_m = reci * KM2M
    veci_m = veci * KM2M
    magr = np.linalg.norm(reci_m)

    # Constants for transformation
    a_ = n * a**2
    b_ = np.sqrt(1 - ag**2 - af**2)
    c_ = 1 + chi**2 + psi**2
    b = 1 / (1 + b_)
    g_ = a_ * b_

    # Compute partial derivatives
    f_ = newton_mean_anomaly(meanlon_m, af, ag)
    x = a * ((1 - ag**2 * b) * np.cos(f_) + af * ag * b * np.sin(f_) - af)
    y = a * ((1 - af**2 * b) * np.sin(f_) + af * ag * b * np.cos(f_) - ag)
    xd = n * a**2 / magr * (af * ag * b * np.cos(f_) - (1 - ag**2 * b) * np.sin(f_))
    yd = n * a**2 / magr * ((1 - af**2 * b) * np.cos(f_) - af * ag * b * np.sin(f_))

    # Equinoctial system components
    p0 = 1 / (1 + chi**2 + psi**2)
    f_vec = p0 * np.array([1 - chi**2 + psi**2, 2 * chi * psi, -2 * fr * chi])
    g_vec = p0 * np.array([2 * fr * chi * psi, fr * (1 + chi**2 - psi**2), 2 * psi])
    w_vec = p0 * np.array([2 * chi, -2 * psi, fr * (1 - chi**2 - psi**2)])

    # Partial derivatives wrt af and ag
    partxaf = ag * b * xd / n + a * y * xd / g_ - a
    partyaf = ag * b * yd / n - a * x * xd / g_
    partxag = -af * b * xd / n + a * y * yd / g_
    partyag = -af * b * yd / n - a * x * yd / g_ - a

    partxdaf = a * xd * yd / g_ - a_ / magr**3 * (a * ag * x / (1 + b_) + x * y / b_)
    partydaf = -a * xd**2 / g_ - a_ / magr**3 * (a * ag * y / (1 + b_) - x**2 / b_)
    partxdag = a * yd**2 / g_ + a_ / magr**3 * (a * af * x / (1 + b_) - y**2 / b_)
    partydag = -a * xd * yd / g_ + a_ / magr**3 * (a * af * y / (1 + b_) + x * y / b_)

    # Initialize transformation matrix
    tm = np.zeros((6, 6))

    # Partials of (rx ry rz vx vy vz) w.r.t. sma
    if use_anom_a:
        p0 = 1 / a
        p1 = -1 / (2 * a)
    else:
        p0 = -2 / (3 * n)
        p1 = 1 / (3 * n)

    tm[:3, 0] = p0 * reci_m
    tm[3:, 0] = p1 * veci_m

    # Partials of (rx ry rz vx vy vz) w.r.t. af and ag
    tm[:3, 1] = partxaf * f_vec + partyaf * g_vec
    tm[3:, 1] = partxdaf * f_vec + partydaf * g_vec
    tm[:3, 2] = partxag * f_vec + partyag * g_vec
    tm[3:, 2] = partxdag * f_vec + partydag * g_vec

    # Partials of (rx ry rz vx vy vz) w.r.t. chi and psi
    p0 = 2 * fr / c_
    tm[:3, 3] = p0 * (psi * (y * f_vec - x * g_vec) - x * w_vec)
    tm[3:, 3] = p0 * (psi * (yd * f_vec - xd * g_vec) - xd * w_vec)
    tm[:3, 4] = p0 * (chi * (x * g_vec - y * f_vec) + y * w_vec)
    tm[3:, 4] = p0 * (chi * (xd * g_vec - yd * f_vec) + yd * w_vec)

    # Partials of (rx ry rz vx vy vz) w.r.t. mean longitude
    p0 = 1 / n
    p1 = -n * a**3 / magr**3
    tm[:3, 5] = p0 * veci_m
    tm[3:, 5] = p1 * reci_m

    # Final Cartesian covariance
    cartcov = tm @ eqcov @ tm.T

    return cartcov, tm


########################################################################################
# Equinoctial <-> Classical Elements
########################################################################################


def covcl2eq(
    classcov: ArrayLike, classstate: ArrayLike, fr: int, anom: AnomalyType
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from classical to equinoctial elements.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        classcov (array_like): 6x6 Classical orbital elements covariance matrix
                               in m and m/s
        classstate (array_like): 6x1 Classical orbital elements in km and radians
                                (a, ecc, incl, node, argp, nu or m)
        fr (int): Retrograde factor (+1 or -1)
        anom (AnomalyType): Anomaly type (MEAN_A, TRUE_A, MEAN_N, TRUE_N)

    Returns:
        tuple: (eqcov, tm)
            eqcov (np.ndarray): 6x6 Equinoctial covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """
    # Parse the orbit state
    a, ecc, incl, omega, argp = classstate[:5]
    a *= KM2M

    # Initialize transformation matrix
    tm = np.zeros((6, 6))

    # Partials of a/n wrt (a, ecc, incl, node, argp, nu/M)
    if anom in {AnomalyType.TRUE_A, AnomalyType.MEAN_A}:
        tm[0, 0] = 1
    else:
        tm[0, 0] = -(3 * np.sqrt(MUM / a**3)) / (2 * a)

    # Partials of af wrt (a, ecc, incl, node, argp, nu/M)
    tm[1, 1] = np.cos(fr * omega + argp)
    tm[1, 3] = -ecc * fr * np.sin(fr * omega + argp)
    tm[1, 4] = -ecc * np.sin(fr * omega + argp)

    # Partials of ag wrt (a, ecc, incl, node, argp, nu/M)
    tm[2, 1] = np.sin(fr * omega + argp)
    tm[2, 3] = ecc * fr * np.cos(fr * omega + argp)
    tm[2, 4] = ecc * np.cos(fr * omega + argp)

    # Partials of chi wrt (a, ecc, incl, node, argp, nu/M)
    mult = (0.5 * np.tan(incl * 0.5) ** 2 + 0.5) * fr * np.tan(incl * 0.5) ** (fr - 1)
    tm[3, 2] = np.sin(omega) * mult
    tm[3, 3] = np.tan(incl * 0.5) ** fr * np.cos(omega)

    # Partials of psi wrt (a, ecc, incl, node, argp, nu/M)
    tm[4, 2] = np.cos(omega) * mult
    tm[4, 3] = -np.tan(incl * 0.5) ** fr * np.sin(omega)

    # Partials of meanlonM/meanlonNu wrt (a, ecc, incl, node, argp, nu/M)
    tm[5, 3] = fr
    tm[5, 4] = tm[5, 5] = 1

    # Calculate the output covariance matrix
    eqcov = tm @ classcov @ tm.T

    return eqcov, tm


def coveq2cl(
    eqcov: ArrayLike, eqstate: ArrayLike, fr: int, anom: AnomalyType
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from equinoctial to classical elements.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        eqcov (array_like): 6x6 Equinoctial covariance matrix in m and m/s
        eqstate (array_like): 6x1 Equinoctial orbital state in km and radians
                              (a/n, af, ag, chi, psi, lm/ln)
        fr (int): Retrograde factor (+1 or -1)
        anom (AnomalyType): Anomaly type (MEAN_A, TRUE_A, MEAN_N, TRUE_N)

    Returns:
        tuple: (classcov, tm)
            classcov (np.ndarray): 6x6 Classical covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """
    # Parse the equinoctial state
    if anom in {AnomalyType.TRUE_A, AnomalyType.MEAN_A}:
        a = eqstate[0]
        n = np.sqrt(MUM / a**3)
    else:
        n = eqstate[0]

    # Get the equinoctial elements
    af, ag, chi, psi = eqstate[1:5]

    # Initialize transformation matrix
    tm = np.zeros((6, 6))

    # Partials for semi-major axis or mean motion
    if anom in {AnomalyType.TRUE_A, AnomalyType.MEAN_A}:
        tm[0, 0] = 1
    else:
        tm[0, 0] = -2 / (3 * n) * (MUM / n**2) ** (1 / 3)

    # Partials for eccentricity
    p0 = 1 / np.sqrt(af**2 + ag**2)
    tm[1, 1] = p0 * af
    tm[1, 2] = p0 * ag

    # Partials for inclination
    p1 = 2 * fr / ((1 + chi**2 + psi**2) * np.sqrt(chi**2 + psi**2))
    tm[2, 3] = p1 * chi
    tm[2, 4] = p1 * psi

    # Partials for RAAN (node)
    p2 = 1 / (chi**2 + psi**2)
    tm[3, 3] = p2 * psi
    tm[3, 4] = -p2 * chi

    # Partials for argument of perigee
    p3 = 1 / (af**2 + ag**2)
    tm[4, 1] = -p3 * ag
    tm[4, 2] = p3 * af
    tm[4, 3] = -fr * p2 * psi
    tm[4, 4] = fr * p2 * chi

    # Partials for anomaly
    if anom in {AnomalyType.TRUE_A, AnomalyType.TRUE_N}:
        p4 = 1 / (af**2 + ag**2)
        tm[5, 1] = p4 * ag
        tm[5, 2] = -p4 * af
    else:
        tm[5, 1] = p3 * ag
        tm[5, 2] = -p3 * af

    tm[5, 5] = 1

    # Compute the transformed covariance matrix
    classcov = tm @ eqcov @ tm.T

    return classcov, tm


########################################################################################
# Cartesian <-> Flight Parameters
########################################################################################


def _compute_partials_az(reci_m, veci_m):
    """Compute partial derivatives of azimuth w.r.t. (rx, ry, rz, vx, vy, vz)."""
    rx, ry, rz = reci_m
    vx, vy, vz = veci_m
    magr = np.linalg.norm(reci_m)
    magv = np.linalg.norm(veci_m)
    rdotv = np.dot(reci_m, veci_m)

    # Sal from mathcad methoc
    p2 = 1 / ((magv**2 - (rdotv / magr) ** 2) * (rx**2 + ry**2))
    k1 = np.linalg.norm(reci_m) * np.cross(reci_m[:2], veci_m[:2])
    k2 = ry * (ry * vz - rz * vy) + rx * (rx * vz - rz * vx)
    k12_sq = k1**2 + k2**2

    # Construct the transformation matrix
    tm_az = np.zeros(6)
    tm_az[0] = p2 * (
        vy * (magr * vz - rz * rdotv / magr)
        - (rx * vy - ry * vx) / magr * (rx * vz - rz * vx + rx * rdotv / magr)
    )
    p2 = 1 / (magr * k12_sq)
    tm_az[0] = p2 * (
        k1 * magr * (rz * vx - 2 * rx * vz)
        + k2 * (-ry * vx * rx + vy * rx**2 + vy * magr**2)
    )
    tm_az[1] = p2 * (
        k1 * magr * (rz * vy - 2 * ry * vz)
        + k2 * (rx * vy * ry - vx * ry**2 - vx * magr**2)
    )
    p2 = k1 / (magr**2 * k12_sq)
    tm_az[2] = p2 * (k2 * rz + (rx * vx + ry * vy) * magr**2)
    p2 = 1 / k12_sq
    tm_az[3] = p2 * (k1 * rx * rz - k2 * ry * magr)
    tm_az[4] = p2 * (k1 * ry * rz + k2 * rx * magr)
    tm_az[5] = -p2 * (k1 * (rx**2 + ry**2))

    return tm_az


def covct2fl(
    cartcov: ArrayLike,
    cartstate: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
    use_latlon: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from Cartesian to flight parameters.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        cartcov (array_like): 6x6 Cartesian covariance matrix in m and m/s
        cartstate (array_like): 6x1 Cartesian state vector in km and km/s
                                (rx, ry, rz, vx, vy, vz)
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 nutation parameters
        eqeterms (bool, optional): Add terms for ast calculation (default True)
        use_latlon (bool, optional): Flag to use lat/lon instead of ra/dec
                                     (default True)

    Returns:
        tuple: (flcov, tm)
            flcov (np.ndarray): 6x6 Flight parameters covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """
    # Parse input state into cartesian components (in meters and m/s)
    reci_m = np.array(cartstate[:3]) * KM2M
    veci_m = np.array(cartstate[3:]) * KM2M

    # Convert to ECEF coordinates if using lat/lon
    r = reci_m
    if use_latlon:
        aeci = np.array([0, 0, 0])
        recef, *_ = fc.eci2ecef(
            reci_m / KM2M,
            veci_m / KM2M,
            aeci,
            ttt,
            jdut1,
            lod,
            xp,
            yp,
            ddpsi,
            ddeps,
            iau80arr,
            eqeterms,
        )
        r = recef * KM2M

    # Calculate common quantities
    magr = np.linalg.norm(reci_m)
    magv = np.linalg.norm(veci_m)
    h = np.linalg.norm(np.cross(reci_m, veci_m))

    # Initialize transformation matrix
    tm = np.zeros((6, 6))

    # Transformation matrix components for latlon or radec
    r_xy2 = np.sum(r[:2] ** 2)
    r_xy = np.sqrt(r_xy2)

    p0 = 1 / r_xy2
    tm[0, 0:2] = [-p0 * r[1], p0 * r[0]]

    p0 = 1 / (magr**2 * r_xy)
    tm[1, 0:2] = -p0 * r[:2] * r[2]
    tm[1, 2] = r_xy / magr**2

    # Partial of flight path angle (fpa) wrt (x, y, z, vx, vy, vz)
    p0 = 1 / (magr**2 * h)
    p1 = 1 / (magv**2 * h)
    tm[2, :3] = p0 * (veci_m * np.sum(reci_m**2) - reci_m * np.sum(reci_m * veci_m))
    tm[2, 3:] = p1 * (reci_m * np.sum(veci_m**2) - veci_m * np.sum(reci_m * veci_m))

    # Partial of azimuth (az) wrt (x, y, z, vx, vy, vz)
    tm[3, :] = _compute_partials_az(reci_m, veci_m)

    # Partial of r and v wrt (x, y, z, vx, vy, vz)
    tm[4, :3] = reci_m / magr
    tm[5, 3:] = veci_m / magv

    # Calculate the output covariance matrix
    flcov = tm @ cartcov @ tm.T

    return flcov, tm


def covfl2ct(
    flcov: ArrayLike,
    flstate: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
    use_latlon: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 covariance matrix from flight parameters to Cartesian elements.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        flcov (array_like): 6x6 Flight parameters covariance matrix in m and m/s
        flstate (array_like): 6x1 Flight parameters state vector in km, km/s and radians
                              (lon or ra, latgc or dec, fpa, az, r, v)
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 nutation parameters
        eqeterms (bool, optional): Add terms for ast calculation (default True)
        use_latlon (bool, optional): Flag to use lat/lon instead of ra/dec
                                     (default True)

    Returns:
        tuple: (cartcov, tm)
            cartcov (np.ndarray): 6x6 Cartesian covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix

    TODO:
        - The returned covariance for lat/lon seems incorrect - investigate
    """
    # Parse input flight state
    lon, latgc, fpa, az, magr, magv = flstate
    magr *= KM2M  # convert to meters
    magv *= KM2M

    # fpa/az trigonometric terms
    cfpa, sfpa = np.cos(fpa), np.sin(fpa)
    caz, saz = np.cos(az), np.sin(az)

    # Initialize transformation matrix
    tm = np.zeros((6, 6))

    # Decide which coordinates to use
    craf, sraf, cdf, sdf = 0, 0, 0, 0
    if use_latlon:
        # Compute ECEF coordinates
        craf, sraf = np.cos(lon), np.sin(lon)
        cdf, sdf = np.cos(latgc), np.sin(latgc)

        # Get ECEF vectors in km and km/s
        recef = np.array([magr * cdf * craf, magr * cdf * sraf, magr * sdf]) / KM2M
        vecef = (
            magv
            / KM2M
            * np.array(
                [
                    -craf * sdf * caz * cfpa - sraf * saz * cfpa + craf * cdf * sfpa,
                    -sraf * sdf * caz * cfpa + craf * saz * cfpa + sraf * cdf * sfpa,
                    sdf * sfpa + cdf * caz * cfpa,
                ]
            )
        )

        # Convert to ECI
        aecef = np.zeros(3)
        reci, veci, _ = fc.ecef2eci(
            recef,
            vecef,
            aecef,
            ttt,
            jdut1,
            lod,
            xp,
            yp,
            ddpsi,
            ddeps,
            iau80arr,
            eqeterms,
        )
        reci *= KM2M  # now in meters
        veci *= KM2M

        # Compute RA/dec from ECI vectors
        temp = np.sqrt(reci[0] ** 2 + reci[1] ** 2)
        rtasc = (
            np.arctan2(reci[1], reci[0])
            if temp >= SMALL
            else np.arctan2(veci[1], veci[0])
        )
        decl = np.arcsin(reci[2] / magr)
    else:
        # Use RA/dec directly
        rtasc, decl = lon, latgc

    # ra/dec trigonometric terms
    cra, sra = np.cos(rtasc), np.sin(rtasc)
    cd, sd = np.cos(decl), np.sin(decl)

    # Reusable variables
    if use_latlon:
        s1, c1 = sraf, craf
        s2, c2 = sdf, cdf
    else:
        s1, c1 = sra, cra
        s2, c2 = sd, cd

    # Position partials w.r.t. (lon latgc fpa az r v)
    tm[0, :2] = [-magr * c2 * s1, -magr * s2 * c1]  # rx
    tm[0, 4] = cd * cra
    tm[1, :2] = [magr * c2 * c1, -magr * s2 * s1]  # ry
    tm[1, 4] = cd * sra
    tm[2, 1] = magr * c2  # rz
    tm[2, 4] = sd

    # Partial of vx wrt (lon, latgc, fpa, az, r, v)
    tm[3, 0] = -magv * (-s1 * caz * s2 * cfpa + c1 * saz * cfpa + c2 * s1 * sfpa)
    tm[3, 1] = -c1 * magv * (s2 * sfpa + c2 * caz * cfpa)
    tm[3, 2] = magv * (cra * caz * sd * sfpa + sra * saz * sfpa + cd * cra * cfpa)
    tm[3, 3] = magv * (cra * saz * sd * cfpa - sra * caz * cfpa)
    tm[3, 5] = -cra * caz * sd * cfpa - sra * saz * cfpa + cd * cra * sfpa

    # Partial of vy wrt (lon, latgc, fpa, az, r, v)
    tm[4, 0] = magv * (-c1 * caz * s2 * cfpa - s1 * saz * cfpa + c2 * c1 * sfpa)
    tm[4, 1] = -s1 * magv * (s2 * sfpa + c2 * caz * cfpa)
    tm[4, 2] = magv * (sra * caz * sd * sfpa - cra * saz * sfpa + cd * sra * cfpa)
    tm[4, 3] = magv * (sra * saz * sd * cfpa + cra * caz * cfpa)
    tm[4, 5] = -sra * caz * sd * cfpa + cra * saz * cfpa + cd * sra * sfpa

    # Partial of vz wrt (lon, latgc, fpa, az, r, v)
    tm[5, 1] = magv * (c2 * sfpa - s2 * caz * cfpa)
    tm[5, 2] = magv * (sd * cfpa - cd * caz * sfpa)
    tm[5, 3] = -magv * cd * saz * cfpa
    tm[5, 5] = sd * sfpa + cd * caz * cfpa

    # Compute Cartesian covariance
    cartcov = tm @ flcov @ tm.T

    return cartcov, tm


########################################################################################
# Satellite Coordinate Systems
########################################################################################


def covct2rsw(
    cartcov: ArrayLike, cartstate: ArrayLike
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 Cartesian covariance matrix to an orbit plane RSW frame.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        cartcov (array_like): 6x6 Cartesian covariance matrix in (m, m/s) or (km, km/s)
        cartstate (array_like): 6x1 Cartesian state vector in km and km/s
                                (rx, ry, rz, vx, vy, vz)

    Returns:
        tuple: (covrsw, tm)
            covrsw (np.ndarray): 6x6 RSW covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """
    # Extract position and velocity vectors
    r = np.array(cartstate[:3])
    v = np.array(cartstate[3:])

    # Define RSW unit vectors
    rv = unit(r)  # along the position vector (radial direction)
    temv = np.cross(r, v)
    wv = unit(temv)  # along the angular momentum vector (out of plane)
    sv = np.cross(wv, rv)  # along the direction perpendicular to rv and wv (in-plane)

    # Initialize the transformation matrix
    tm = np.zeros((6, 6))

    # Populate the position transformation submatrix
    tm[0, 0:3] = rv
    tm[1, 0:3] = sv
    tm[2, 0:3] = wv

    # Populate the velocity transformation submatrix
    tm[3, 3:6] = rv
    tm[4, 3:6] = sv
    tm[5, 3:6] = wv

    # Transform the covariance matrix
    covrsw = tm @ cartcov @ tm.T

    return covrsw, tm


def covct2ntw(
    cartcov: ArrayLike, cartstate: ArrayLike
) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a 6x6 Cartesian covariance matrix to an orbit plane NTW frame.

    References:
        Vallado and Alfano 2015, AAS 15-537

    Args:
        cartcov (array_like): 6x6 Cartesian covariance matrix in (m, m/s) or (km, km/s)
        cartstate (array_like): 6x1 Cartesian state vector in km and km/s
                                (rx, ry, rz, vx, vy, vz)

    Returns:
        tuple: (covntw, tm)
            covntw (np.ndarray): 6x6 NTW covariance matrix in m and m/s
            tm (np.ndarray): 6x6 Transformation matrix
    """
    # Extract position and velocity vectors
    r_eci_m = np.array(cartstate[:3])
    v_eci_m = np.array(cartstate[3:])

    # Define NTW unit vectors
    tv = unit(v_eci_m)  # along the velocity vector
    temv = np.cross(r_eci_m, v_eci_m)
    wv = unit(temv)  # along the angular momentum vector
    nv = np.cross(tv, wv)  # normal to both tv and wv

    # Initialize the transformation matrix
    tm = np.zeros((6, 6))

    # Position transformation submatrix
    tm[0, 0:3] = nv
    tm[1, 0:3] = tv
    tm[2, 0:3] = wv

    # Velocity transformation submatrix
    tm[3, 3:6] = nv
    tm[4, 3:6] = tv
    tm[5, 3:6] = wv

    # Transform the covariance matrix
    covntw = tm @ cartcov @ tm.T

    return covntw, tm



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from typing import Tuple

import numpy as np

from .. import constants as const


def safe_sqrt(value: float, context: str = "") -> float:
    """Safe square root function that checks for negative values.

    Args:
        value (float): The value to take the square root of
        context (str, optional): A description of the current step or variable
                                 being used

    Returns:
        float: The square root of the value

    Raises:
        ValueError: If the value is negative
    """
    if value < 0:
        error_message = f"Cannot take square root of negative value: ({value})"
        if context:
            error_message += f"\nContext: {context}"
        raise ValueError(error_message)
    return np.sqrt(value)


def hms2sec(hours: int, minutes: int, seconds: float) -> float:
    """Convert hours, minutes, and seconds to seconds.

    Args:
        hours (int): The number of hours
        minutes (int): The number of minutes
        seconds (float): The number of seconds

    Returns:
        float: The total number of seconds
    """
    return hours * const.HR2SEC + minutes * const.MIN2SEC + seconds


def sec2hms(seconds: float) -> Tuple[int, int, float]:
    """Convert seconds to hours, minutes, and seconds.

    Args:
        seconds (float): The total number of seconds

    Returns:
        tuple: (hours, minutes, seconds)
            hours (int): The number of hours
            minutes (int): The number of minutes
            seconds (float): The number of seconds
    """
    # Get the hours and the fraction of hours
    total_hours = seconds / const.HR2SEC
    hours = int(total_hours)
    hours_fraction = total_hours - hours

    # Get the minutes and seconds
    minutes = int(hours_fraction * const.MIN2SEC)
    secs = (hours_fraction - minutes / const.MIN2SEC) * const.HR2SEC

    # Adjust seconds to avoid floating point errors
    secs = round(secs) if abs(secs - round(secs)) < const.SMALL else secs

    return hours, minutes, secs


def hms2ut(hours: int, minutes: int, seconds: float) -> float:
    """Converts hours, minutes, and seconds into universal time.

    Args:
        hours (int): Hours (0 .. 23)
        minutes (int): Minutes (0 .. 59)
        seconds (float): Seconds (0.0 .. 59.999)

    Returns:
        float: Universal time in hrmin.sec format
    """
    return hours * 100 + minutes + seconds * 0.01


def ut2hms(ut: float) -> Tuple[int, int, float]:
    """Converts universal time (hhmm.sec format) into hours, minutes, and seconds.

    Args:
        ut (float): Universal time in hrmin.sec format

    Returns:
        tuple: (hours, minutes, seconds)
            hours (int): The number of hours (0 .. 23)
            minutes (int): The number of minutes (0 .. 59)
            seconds (float): The number of seconds (0.0 .. 59.999)
    """
    hr = int(np.floor(ut * 0.01))
    minute = int(np.floor(ut - hr * 100))
    second = (ut - hr * 100 - minute) * 100
    return hr, minute, second


def hms2rad(hours: int, minutes: int, seconds: float) -> float:
    """Convert hours, minutes, and seconds to radians.

    References:
        Vallado: 2022, p. 199, Algorithm 19

    Args:
        hours (int): The number of hours
        minutes (int): The number of minutes
        seconds (float): The number of seconds

    Returns:
        float: The total number of radians
    """
    return (hours + minutes / const.MIN2SEC + seconds / const.HR2SEC) * const.HR2RAD


def rad2hms(radians: float) -> Tuple[int, int, float]:
    """Convert radians to hours, minutes, and seconds.

    References:
        Vallado: 2022, p. 199-200, Algorithm 20

    Args:
        radians (float): The total number of radians

    Returns:
        tuple: (hours, minutes, seconds)
            hours (int): The number of hours
            minutes (int): The number of minutes
            seconds (float): The number of seconds
    """
    # Get the total seconds from the radians
    total_seconds = radians / const.HR2RAD * const.HR2SEC

    # Convert the total seconds to hours, minutes, and seconds
    return sec2hms(total_seconds)


def dms2rad(degrees: int, minutes: int, seconds: float) -> float:
    """Convert degrees, minutes, and seconds to radians.

    References:
        Vallado: 2022, p. 198, Algorithm 17

    Args:
        degrees (int): The number of degrees
        minutes (int): The number of minutes
        seconds (float): The number of seconds

    Returns:
        float: The total number of radians
    """
    return float(
        np.radians(degrees + minutes / const.DEG2MIN + seconds / const.DEG2ARCSEC)
    )


def rad2dms(radians: float) -> Tuple[int, int, float]:
    """Convert radians to degrees, minutes, and seconds.

    References:
        Vallado: 2022, p. 199, Algorithm 18

    Args:
        radians (float): The total number of radians

    Returns:
        tuple: (degrees, minutes, seconds)
            degrees (int): The number of degrees
            minutes (int): The number of minutes
            seconds (float): The number of seconds
    """
    # Convert radians to total degrees
    total_degrees = np.degrees(radians)
    degrees = int(total_degrees)
    degrees_fraction = total_degrees - degrees

    # Get minutes and seconds
    minutes = int(degrees_fraction * const.DEG2MIN)
    secs = (degrees_fraction - minutes / const.DEG2MIN) * const.DEG2ARCSEC

    # Adjust seconds to avoid floating-point errors
    secs = round(secs) if abs(secs - round(secs)) < const.SMALL else secs

    return degrees, minutes, secs



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import calendar
from datetime import datetime, timedelta
from typing import Dict, List, Tuple

from .. import constants as const


# Constants
SEC2MICROSEC = 1e6


def initialize_time(year: int) -> Dict[str, List]:
    """Initializes time-related information for a given year.

    Args:
        year (int): The year for which the data is initialized
                    (to account for leap years)

    Returns:
        dict[str, List]: A dictionary with three keys:
              - 'lmonth': List of days in each month for the given year
              - 'monthtitle': List of abbreviated month names
              - 'daytitle': List of abbreviated day names (starting with Sunday)
    """
    return {
        "lmonth": [calendar.monthrange(year, i)[1] for i in range(1, 13)],
        "monthtitle": [calendar.month_abbr[i].lower() for i in range(1, 13)],
        "daytitle": [calendar.day_abbr[(i + 6) % 7].lower() for i in range(7)],
    }


def get_int_month(month_str: str):
    """Converts a 3-character month abbreviation to its integer equivalent.

    Args:
        month_str (str): Abbreviation of the month (e.g., 'jan', 'feb')

    Returns:
        int: Integer representation of the month (1 for January, 2 for February, etc.)

    Raises:
        ValueError: If the input string is not a valid month abbreviation.
    """
    try:
        return datetime.strptime(month_str[:3].capitalize(), "%b").month
    except ValueError as e:
        raise ValueError(f"Invalid month abbreviation: {month_str}") from e


def get_int_day(day_str: str) -> int:
    """Converts a 3-character day abbreviation to its integer equivalent.

    Args:
        day_str (str): Abbreviation of the day (e.g., 'sun', 'mon')

    Returns:
        int: Integer representation of the day (1 for Sunday, 2 for Monday, etc.)

    Raises:
        ValueError: If the day abbreviation is invalid
    """
    # Shift days to start with Sunday
    shifted_days = [calendar.day_abbr[(i + 6) % 7].lower() for i in range(7)]
    try:
        return shifted_days.index(day_str.lower()) + 1
    except ValueError as e:
        raise ValueError(f"Invalid day abbreviation: {day_str}") from e


def days2mdh(year: int, days: float) -> Tuple[int, int, int, int, float]:
    """Converts day of the year to month, day, hour, minute, and second.

    Args:
        year (int): Year (e.g., 1900 .. 2100)
        days (float): Day of the year (1.0 .. 366.0)

    Returns:
        tuple: (month, day, hour, minute, second)
            month (int): Integer month (1 .. 12)
            day (int): Integer day (1 .. 31)
            hour (int): Integer hour (0 .. 23)
            minute (int): Integer minute (0 .. 59)
            second (float): Seconds (0.0 .. 59.999999)
    """
    # Get the start of the year
    start_of_year = datetime(year, 1, 1)

    # Calculate the full date and time by adding the fractional days
    delta = timedelta(days=days - 1)  # subtract 1 since days=1.0 means Jan 1
    full_date = start_of_year + delta

    # Extract the components
    month = full_date.month
    day = full_date.day
    hour = full_date.hour
    minute = full_date.minute
    second = full_date.second + full_date.microsecond / SEC2MICROSEC

    return month, day, hour, minute, second


def find_days(
    year: int, month: int, day: int, hour: int, minute: int, second: float
) -> float:
    """Finds the fractional days through a year.

    References:
        Vallado: 2022, p. 202

    Args:
        year (int): Year (1900 .. 2100)
        month (int): Month (1 .. 12)
        day (int): Day (1 .. 28, 29, 30, 31)
        hour (int): Hour (0 .. 23)
        minute (int): Minute (0 .. 59)
        second (float): Second (0.0 .. 59.999)

    Returns:
        float: Day of the year plus fraction of the day
    """
    # Start of the year
    start_of_year = datetime(year, 1, 1)

    # Current date and time
    current_date = datetime(
        year, month, day, hour, minute, int(second), int((second % 1) * SEC2MICROSEC)
    )

    # Compute the difference in days
    delta = current_date - start_of_year

    # Fractional days
    return (
        delta.days
        + delta.seconds / const.DAY2SEC
        + delta.microseconds / const.DAY2SEC / SEC2MICROSEC
    ) + 1  # add 1 since days=1.0 means Jan 1



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import numpy as np
from numpy.typing import ArrayLike

from ..constants import SMALL


########################################################################################
# Axes Rotations
########################################################################################


def rot1(vec: ArrayLike, xval: float) -> np.ndarray:
    """Rotation about the 1st axis (x-axis)

    Args:
        vec (array_like): Input vector
        xval (float): Angle of rotation in radians

    Returns:
        np.ndarray: Rotated vector
    """
    c, s = np.cos(xval), np.sin(xval)
    return np.array([vec[0], c * vec[1] + s * vec[2], c * vec[2] - s * vec[1]])


def rot2(vec: ArrayLike, xval: float) -> np.ndarray:
    """Rotation about the 2nd axis (y-axis)

    Args:
        vec (array_like): Input vector
        xval (float): Angle of rotation in radians

    Returns:
        np.ndarray: Rotated vector
    """
    c, s = np.cos(xval), np.sin(xval)
    return np.array([c * vec[0] - s * vec[2], vec[1], c * vec[2] + s * vec[0]])


def rot3(vec: ArrayLike, xval: float) -> np.ndarray:
    """Rotation about the 3rd axis (z-axis)

    Args:
        vec (array_like): Input vector
        xval (float): Angle of rotation in radians

    Returns:
        np.ndarray: Rotated vector
    """
    c, s = np.cos(xval), np.sin(xval)
    return np.array([c * vec[0] + s * vec[1], c * vec[1] - s * vec[0], vec[2]])


########################################################################################
# Rotation Matrices
########################################################################################


def rot1mat(xval: float) -> np.ndarray:
    """Rotation matrix for an input angle about the first axis.

    Args:
        xval (float): Angle of rotation in radians

    Returns:
        np.ndarray: Rotation matrix
    """
    c, s = np.cos(xval), np.sin(xval)
    return np.array([[1, 0, 0], [0, c, s], [0, -s, c]])


def rot2mat(xval: float) -> np.ndarray:
    """Rotation matrix for an input angle about the second axis.

    Args:
        xval (float): Angle of rotation in radians

    Returns:
        np.ndarray: Rotation matrix
    """
    c, s = np.cos(xval), np.sin(xval)
    return np.array([[c, 0, -s], [0, 1, 0], [s, 0, c]])


def rot3mat(xval: float) -> np.ndarray:
    """Rotation matrix for an input angle about the third axis.

    Args:
        xval (float): Angle of rotation in radians

    Returns:
        np.ndarray: Rotation matrix
    """
    c, s = np.cos(xval), np.sin(xval)
    return np.array([[c, s, 0], [-s, c, 0], [0, 0, 1]])


########################################################################################
# Vector Math
########################################################################################


def angle(v1: ArrayLike, v2: ArrayLike) -> float:
    """Calculates the angle between two vectors in radians

    This function computes the angle between two vectors using the dot product
    and the magnitudes of the vectors. The function handles cases where the dot
    product might slightly exceed the interval [-1, 1] due to numerical
    precision issues by clipping the value. If either vector has zero
    magnitude, the function returns `np.nan`to indicate that the angle is not
    computable.

    Args:
        v1 (array_like): The first vector
        v2 (array_like): The second vector

    Returns:
        float: The angle between the two vectors in radians (returns `np.nan`
               if either vector is zero)
    """
    mag_v1 = np.linalg.norm(v1)
    mag_v2 = np.linalg.norm(v2)

    if mag_v1 * mag_v2 > SMALL**2:
        cos_angle = np.dot(v1, v2) / (mag_v1 * mag_v2)
        cos_angle = np.clip(cos_angle, -1, 1)  # keep cosine within domain
        return np.arccos(cos_angle)
    else:
        return np.nan


def unit(v: ArrayLike) -> np.ndarray:
    """Returns the unit vector of a given vector

    Args:
        v (array_like): The input vector

    Returns:
        numpy.ndarray: The unit vector of the input vector
                       (v / ||v|| if ||v|| > 0, 0 otherwise)
    """
    mag = np.linalg.norm(v)
    return v / mag if mag > SMALL else np.zeros_like(v)



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 15 Jan 2008
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike


def finite_diff(
    pertelem: int,
    percentchg: float,
    deltaamtchg: float,
    xnom: ArrayLike,
    growth_factor: float = 1.4,
) -> Tuple[float, np.ndarray]:
    """Perturbs the components of the state vector for finite differencing.

    References:
        Vallado: 2022, p. 779-780

    Args:
        pertelem (int): Index of the element to perturb (0-based)
        percentchg (float): Amount to modify the vector by in finite differencing
        deltaamtchg (float): Tolerance for small value in finite differencing
        xnom (np.ndarray): State vector to perturb
        growth_factor (float, optional): Factor to increase perturbation by
                                         (defaults to 1.4)

    Returns:
        tuple (deltaamt, xnomp):
            deltaamt (float): Amount of perturbation
            xnomp (np.ndarray): Perturbed state vector
    """
    deltaamt = 0
    xnomp = xnom.copy()

    for _ in range(len(xnom)):
        deltaamt = xnom[pertelem] * percentchg
        xnomp[pertelem] += deltaamt

        if abs(deltaamt) >= deltaamtchg:
            break  # perturbation is sufficient
        else:
            percentchg *= growth_factor  # increase perturbation by growth factor

    return deltaamt, xnomp
